/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file DiscreteDimension.h
 * Header definition of the class DiscreteDimension.
 *
 * @author A. Huber
 * @since 27.09.2018
 */

#ifndef DISKRDIMENSIONEN_H
#define DISKRDIMENSIONEN_H

#include "../problem.h"
#include "DiscreteDefs.h"

namespace OCPbasic
{

    template <typename real>
    class DiscreteDimension
    {
      public:
        /**
         * \brief DiscreteDimension:
         *        Constructor of the class DiscreteDimension.
         *
         * \param[in] P     Optimal controll problem.
         * \param[in] INFTY Value of \f$\infty\f$.
         * \param[in] TOL   Tolerance to \f$0\f$.
         *
         */
        DiscreteDimension(Problem<real> * P, int_type N, real INFTY, real TOL);

        /**
         * \brief ~DiscreteDimension:
         *        Virtual destructor of the class DiscreteDimension.
         */
        virtual ~DiscreteDimension();

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                   public Memberfunktionen                                                           //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief getBoundTypeArrayNLCSTR:
         *        Returns the pointer of the identification array of the nonlinear constraints.
         *
         * @return Returns the pointer of the identification array of the nonlinear constraints.
         *
         */
        DiscreteDefs::BoundType const & getBoundTypeArrayNLCSTR(int_type idx) const
        {
            return this->BoundTypeArrayNLCSTR[size_t(idx)];
        }

        /**
         * \brief getBoundTypeArrayBOUND:
         *        Returns the pointer of the identification array of the boundary constraints.
         *
         * @return Returns the pointer of the identification array of the boundary constraints.
         *
         */
        DiscreteDefs::BoundType const & getBoundTypeArrayBOUND(int_type idx) const
        {
            return this->BoundTypeArrayBOUND[size_t(idx)];
        }

        /**
         * \brief getBoundTypeArraySTATE:
         *        Returns the pointer of the identification array of the state constraints.
         *
         * @return Returns the pointer of the identification array of the state constraints.
         *
         */
        DiscreteDefs::BoundType const & getBoundTypeArraySTATE(int_type idx) const
        {
            return this->BoundTypeArraySTATE[size_t(idx)];
        }

        /**
         * \brief getBoundTypeArrayCONTROL:
         *        Returns the pointer of the identification array of the controll constraints.
         *
         * @return Array zur Identifikation der verschiedenen Steuerungsbeschränkungen.
         *
         */
        DiscreteDefs::BoundType const & getBoundTypeArrayCONTROL(int_type idx) const
        {
            return this->BoundTypeArrayCONTROL[size_t(idx)];
        }

        /**
         * \brief getBoundTypeArrayPARAM:
         *        Returns the pointer of the identification array of the parameter constraints.
         *
         * @return Returns the pointer of the identification array of the parameter constraints.
         *
         */
        DiscreteDefs::BoundType const & getBoundTypeArrayPARAM(int_type idx) const
        {
            return this->BoundTypeArrayPARAM[size_t(idx)];
        }

        /**
         * \brief getBoundDimNLCSTR:
         *        Return the dimensions of the different types of nonlinear constraints.
         *
         * @return Return the dimensions of the different types of nonlinear constraints.
         *
         */
        DiscreteDefs::BoundDimension const & getBoundDimNLCSTR() const
        {
            return this->BoundDimNLCSTR;
        }

        /**
         * \brief getBoundDimBOUND:
         *        Return the dimensions of the different types of boundary constraints.
         *
         * @return Return the dimensions of the different types of boundary constraints.
         *
         */
        DiscreteDefs::BoundDimension const & getBoundDimBOUND() const
        {
            return this->BoundDimBOUND;
        }

        /**
         * \brief getBoundDimSTATE:
         *        Return the dimensions of the different types of state constraints.
         *
         * @return Return the dimensions of the different types of state constraints.
         *
         */
        DiscreteDefs::BoundDimension const & getBoundDimSTATE() const
        {
            return this->BoundDimSTATE;
        }

        /**
         * \brief getBoundDimCONTROL:
         *        Return the dimensions of the different types of controll constraints.
         *
         * @return Return the dimensions of the different types of controll constraints.
         *
         */
        DiscreteDefs::BoundDimension const & getBoundDimCONTROL() const
        {
            return this->BoundDimCONTROL;
        }

        /**
         * \brief getBoundDimPARAM:
         *        Return the dimensions of the different types of parameter constraints.
         *
         * @return Return the dimensions of the different types of parameter constraints.
         *
         */
        DiscreteDefs::BoundDimension const & getBoundDimPARAM() const
        {
            return this->BoundDimPARAM;
        }

        /**
         * \brief set_lambda_hesse:
         *        Calculates the resulting lagrange multipliers of the hessian.
         *
         * \param[in]  lambda_IP        Lagrange multipliers of the interior-point-mathod.
         * \param[out] lambda_hesse     Lagrange multipliers of the hessian.
         *
         */
        void set_lambda_hesse(real const lambda_IP[], real lambda_hesse[]);

        /**
         * \brief get_ParameterConstraintsMatrix:
         *        Calculates the jacobian of the parameter constraints.
         *
         * \param[out] Q    Jacobian of the parameter constraints.
         *
         * @return True, if dimensions are ok.
         *
         */
        bool get_ParameterConstraintsMatrix(MatrixWrapper<real> & Q);

        /**
         * \brief getDimLambdaIP:
         *        Returns dimension of the lagrange multipliers of the interior-point-method.
         *
         * @return Returns dimension of the lagrange multipliers of the interior-point-method.
         *
         */
        int_type getDimLambdaIP();

        /**
         * \brief getDimSIP:
         *        Returns dimension of the slack variables of the interior-point-method.
         *
         * @return Returns dimension of the slack variables of the interior-point-method.
         *
         */
        int_type getDimSIP();

        /**
         * \brief set_C_Constr:
         *        Calculates the submatrix of the constraints from an original matrix.
         *
         * \param[in]  G         Original jacobian matrix.
         * \param[out] C_        New jacobian matrix for the interior-point-method.
         *
         */
        void set_C_Constr(MatrixWrapper<real> & G, MatrixWrapper<real> & C_);

        /**
         * \brief set_Q_Constr:
         *        Calculates the submatrix of the parameter constraints from an original matrix.
         *
         * \param[in]  q    Original parameter jacobian matrix.
         * \param[out] Q    New parameter jacobian matrix for the interior-point-method.
         *
         */
        void set_Q_Constr(MatrixWrapper<real> & q, MatrixWrapper<real> & Q);

        /**
         * \brief set_PSI:
         *        Calculates the submatrix of the boundary constraints from an original matrix.
         *
         * \param[in]  psi  Original boundary jacobian matrix.
         * \param[out] PSI  New boundary jacobian matrix for the interior-point-method.
         *
         */
        void set_PSI(MatrixWrapper<real> & psi, MatrixWrapper<real> & PSI);

        /**
         * \brief set_Vector_C_from_G:
         *        Transforms the vector of the nonlinear constraints into the constraint vector of the
         *        interior-point-method.
         *
         * \param[in]  g    Vector of the nonlinear constraints
         *                  \f$g_{\mathrm{min} \leq g \leq g_{\mathrm{max}}\f$.
         * \param[in]  y_k  State vector of index \f$k\f$.
         * \param[in]  u_k  Control vector of index \f$k\f$.
         * \param[in]  P    Optimal control problem.
         * \param[out] c    Vector of the nonlinear constraints of the interior-point-method
         *                  c \leq 0\f$.
         *
         */
        void set_Vector_C_from_G(real const g[], real const y_k[], real const u_k[], Problem<real> const * P, real c[]);

        /**
         * \brief set_Vector_IPbound_from_bound:
         *        Transforms the vector of the boundary constraints into the boundary constraint vector of the
         *        interior-point-method.
         *
         * \param[in]  bound        Boundary constraints
         *                          \f$\psi_{\mathrm{min} \leq \psi \leq \psi_{\mathrm{max}}\f$.
         * \param[in]  P            Optimal control problem.
         * \param[out] OUTBound     boundary constraints of the interior-point-method.
         *                          \f$\hat{\psi} \leq 0\f$.
         *
         */
        void set_Vector_IPbound_from_bound(real const bound[], Problem<real> const * P, real OUTBound[]);

        /**
         * \brief get_isEquVec:
         *        Sets the isEqu vector.
         *
         * \param[out] isEquVec  Is equality vector.
         *
         */
        void get_isEquVec(bool isEquVec[]);

      private:
        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                  private Membervariablen                                                            //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        IPbasic::Malloc<real> mem_real;

        /// Variable of \f$\infty\f$.
        real INF;
        /// Variable of the toerance to zero.
        real TOL;

        /// Number of gridpoints.
        int_type N;
        /// Dimension of the lagrange multipliers in the interior-point-method.
        int_type DimLambdaIP;
        /// Dimension of the slack variables in the interior-point-method.
        int_type Dim_Schlupf;
        /// Pointer to a char vector, which characterizes the inequalities.
        std::vector<uint8_t> Case_lambda_k;

        /// Dimensions of the nonlinear constraints.
        DiscreteDefs::BoundDimension BoundDimNLCSTR;
        /// Dimensions of the boundary constraints.
        DiscreteDefs::BoundDimension BoundDimBOUND;
        /// Dimensions of the state constraints.
        DiscreteDefs::BoundDimension BoundDimSTATE;
        /// Dimensions of the control constraints.
        DiscreteDefs::BoundDimension BoundDimCONTROL;
        /// Dimensions of the parameter constraints.
        DiscreteDefs::BoundDimension BoundDimPARAM;

        /// Identifikation array of the different nonlinear constraints.
        std::vector<DiscreteDefs::BoundType> BoundTypeArrayNLCSTR;
        /// Identifikation array of the different boundary constraints.
        std::vector<DiscreteDefs::BoundType> BoundTypeArrayBOUND;
        /// Identifikation array of the different state constraints.
        std::vector<DiscreteDefs::BoundType> BoundTypeArraySTATE;
        /// Identifikation array of the different control constraints.
        std::vector<DiscreteDefs::BoundType> BoundTypeArrayCONTROL;
        /// Identifikation array of the different parameter constraints.
        std::vector<DiscreteDefs::BoundType> BoundTypeArrayPARAM;

        /// Indexing of the nonlinear constraints.
        std::vector<int_type> Index_NLCSTR;
        /// Indexing of the boundary constraints.
        std::vector<int_type> Index_BOUND;

        /// Jacobian of the states. (pointwise)
        MatrixWrapper<real> C_Constr_Y;
        /// Jacobian of the controls. (pointwise)
        MatrixWrapper<real> C_Constr_U;
        /// Jacobian of the parameters. (pointwise)
        MatrixWrapper<real> C_Constr_P;

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                  private Memberfunktionen                                                           //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief setIndexArrays:
         *        Sets all the Indexarrays for the nonlinear, boundary, state, control and parameter constraints.
         *
         * \param[in] P     Optimal control problem.
         *
         */
        void setIndexArrays(Problem<real> * P);

        /**
         * \brief set_internal_Constr_YUP:
         *        Sets the internal jacobians of the states, controls and parameters.
         */
        void set_internal_Constr_YUP();

        /**
         * \brief set_internal_linear_Constr:
         *        Calculates a jacobian for linear constraints.
         *
         * \param[in]  BoundDim      Dimensions of the linear constraints.
         * \param[in]  BOUNDTYPE     Array of the types of linear constraints.
         * \param[out] M             Output jacobian.
         *
         */
        void set_internal_linear_Constr(DiscreteDefs::BoundDimension * BoundDim, DiscreteDefs::BoundType const BOUNDTYPE[], MatrixWrapper<real> & M);

        /**
         * \brief set_internal_Index_NLCSTR:
         *        Sets thie index-array of the nonlinear constraints.
         */
        void set_internal_Index_NLCSTR();

        /**
         * \brief set_internal_Index_BOUND:
         *        Sets thie index-array of the boundary constraints.
         */
        void set_internal_Index_BOUND();

        /**
         * \brief set_Q_Index:
         *        Gets the jacobian of the parameters for an index of the parameters constraints.
         *
         * \param[in]  Index    Index of the grid.
         * \param[in]  q        Orginal parameter jacobian.
         * \param[out] Q        New parameter jacobian of the interior-point-method.
         *
         */
        void set_Q_Index(int_type Index, MatrixWrapper<real> & q, MatrixWrapper<real> & Q);

        /**
         * \brief set_CaseLambda:
         *        Sets the vector Case_lambda_k, which is defining the type of inequality.
         */
        void set_CaseLambda();
    };

} // namespace OCPbasic

#endif // DISKRDIMENSIONEN_H
