/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file Discretization.h
 * Header definition of the class Discretization.
 *
 * @author A. Huber
 * @since 21.04.2016
 */

#ifndef DISKRETISIERUNG_H
#define DISKRETISIERUNG_H

#include "../OCP_Params.h"
#include "../OCP_Timings.h"
#include "../linalg/collectormatrix.h"
#include "../problem.h"
#include "../solution.h"
#include "DiscreteDimension.h"
#include <IPbasic/IPbasic.h>

namespace OCPbasic
{

    template <typename real>
    class Discretization
    {
      protected:
        /**
         * \brief Discretization:
         *        Constructor of the class Discretization.
         *
         * \param[in] N     Number of gridpoints in time.
         * \param[in] t0    Startingtime.
         * \param[in] tF    Endtime.
         * \param[in] P     Optimal control problem.
         *
         */
        Discretization(int_type N, real t0, real tF, Problem<real> * P);

      public:
        /**
         * \brief ~Discretization:
         *        Virtual destructor of the class Discretization.
         */
        virtual ~Discretization();

        /// Number of gridpoints in time.
        int_type N;
        /// Startingtime.
        real t0;
        /// Endtime.
        real tF;
        /// Optimal control problem.
        Problem<real> * P;
        /// Hessian of the lagrangian.
        CollectorMatrix<real> * HesseMatrix;
        /// Jacobian of the constraints.
        CollectorMatrix<real> * JacobiMatrix;
        /// Dimensions of the discreet optimal control problem.
        DiscreteDimension<real> * MyDims;
        /// Internal OCP parameters.
        OCPbasic::OCP_Params<real> * MyOCP_Params;
        /// Internal OCP timings.
        OCPbasic::OCP_Timings<real> * MyOCP_Timings;
        /// Local Timer of the Discretization class.
        TicToc LocalTimerDiscretization;

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                     public Memberfunktionen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief regularizeHessian:
         *        Regularize the hessian matrix.
         */
        virtual void regularizeHessian() = 0;

        /**
         * \brief writeToSolution:
         *        Writes the current data into a solution.
         *
         * \param[out] Sol      Solution object.
         *
         * @return True if the dimensions of the solution are correct.
         *
         */
        virtual bool writeToSolution(Solution<real> & Sol) = 0;

        /**
         * \brief readFromSolution:
         *        Reads the data from a solution.
         *
         * \param[in] Sol       Solution object.
         *
         * @return True if the dimensions of the solution are correct.
         *
         */
        virtual bool readFromSolution(Solution<real> & Sol) = 0;

        /**
         * \brief get_x_Vector:
         *        Retruns the current statevector.
         *
         * @return Retruns the current statevector.
         *
         */
        real * get_x_Vector()
        {
            return this->x;
        }

        /**
         * \brief get_Lambda_Vector:
         *        Returns the current vector of the lagrangian multipliers.
         *
         * @return Returns the current vector of the lagrangian multipliers.
         *
         */
        real * get_Lambda_Vector()
        {
            return this->lambda;
        }

        /**
         * \brief getAnzVar:
         *        Returns the dimension of the statevector.
         *
         * @return Returns the dimension of the statevector.
         *
         */

        int_type getAnzVar() const
        {
            return this->N_Statevector;
        }

        /**
         * \brief getAnzNebenBed:
         *        Returns the dimension of the constraints.
         *
         * @return Returns the dimension of the constraints.
         *
         */
        int_type getAnzNebenBed() const
        {
            return this->N_Constraintvector;
        }

        /**
         * \brief getTimeHesse:
         *        Returns the time for computing the hessian.
         *
         * @return Returns the time for computing the hessian.
         *
         */
        real getTimeHesse() const
        {
            return this->TimeHesse;
        }

        /**
         * \brief getTimeJacobi:
         *        Returns the time for computing the jacobian.
         *
         * @return Returns the time for computing the jacobian.
         *
         */
        real getTimeJacobi() const
        {
            return this->TimeJacobi;
        }

        /**
         * \brief getTimeIP:
         *        Returns the time for computing the interior point parts
         *        of the interior-point-method.
         *
         * @return Returns the time for computing the interior point parts
         *         of the interior-point-method.
         *
         */
        real getTimeIP() const
        {
            return this->TimeIP;
        }

        /**
         * \brief getP:
         *        Returns the pointer on the parameter array inside the statevector.
         *
         * @return Returns the pointer on the parameter array.
         *
         */
        virtual real * getP() = 0;

        /**
         * \brief getYk:
         *        Returns the pointer on the state \f$y_k\f$ inside the statevector.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the state \f$y_k\f$.
         *
         */
        virtual real * getYk(int_type Index) = 0;

        /**
         * \brief getUk:
         *        Returns the pointer on the control \f$u_k\f$ inside the statevector.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the control \f$u_k\f$.
         *
         */
        virtual real * getUk(int_type Index) = 0;

        /**
         * \brief getLAMBDAk:
         *        Returns the pointer on the lagrange mutlipliers \f$\lambda_k\f$.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the lagrange mutlipliers \f$\lambda_k\f$.
         *
         */
        virtual real * getLAMBDAk(int_type Index) = 0;

        /**
         * \brief getSIGMA:
         *        Returns the pointer on the lagrange mutlipliers \f$\sigma\f$.
         *
         * @return Returns the pointer on the lagrange mutlipliers \f$\sigma\f$.
         *
         */
        virtual real * getSIGMA() = 0;

        /**
         * \brief getMUk:
         *        Returns the pointer on the lagrange mutlipliers \f$\mu_k\f$.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the lagrange mutlipliers \f$\mu_k\f$.
         *
         */
        virtual real * getMUk(int_type Index) = 0;

        /**
         * \brief getVARPHI:
         *        Retruns the value of the target function for a statevector.
         *
         * \param[in] states        Statevector.
         *
         * @return Retruns the target function for a statevector.
         *
         */
        virtual real getVARPHI(real const states[]) = 0;

        /**
         * \brief Grad_und_VARPHI:
         *        Retruns the value of the target function and its gradient for a statevector.
         *
         * \param[in]  states   Statevector.
         * \param[out] VARPHI   Target function value.
         * \param[out] Grad     Gradient of the target function.
         *
         */
        virtual void Grad_und_VARPHI(real const states[], real VARPHI[], real Grad[]) = 0;

        /**
         * \brief CON_Diskr:
         *        Evaluates the constraints for a statevector.
         *
         * \param[in]  states   Statevector.
         * \param[out] GG       Constraintvector.
         *
         */
        virtual void CON_Diskr(real const states[], real GG[]) = 0;

        /**
         * \brief update_IPData:
         *        Update of the interior point data.
         *
         * \param[in] x             Statevector.
         * \param[in] lambda_IP     Vector of the lagrangian multipliers.
         *
         */
        virtual void update_IPData(real const x[], real const lambda_IP[]) = 0;

        /**
         * \brief getDimStateIP:
         *        Returns the Dimension of the statevector in the IP.
         *
         * @return Returns the Dimension of the statevector in the IP.
         *
         */
        virtual int_type getDimStateIP() = 0;

        /**
         * \brief getDimLambdaIP:
         *        Returns the Dimension of the vector of the lagrangian multipliers in the IP.
         *
         * @return Returns the Dimension of the vector of the lagrangian multipliers in the IP.
         *
         */
        virtual int_type getDimLambdaIP() = 0;

        /**
         * \brief update_Jacobian:
         *        Updates the jacobian for a statevector.
         *
         * \param[in] x             Statevector.
         *
         */
        virtual void update_Jacobian(real const x[]) = 0;

        /**
         * \brief update_Hessian:
         *        Updates the hessian for a statevector and a vector of the lagrangian multipliers.
         *
         * \param[in] x             Statevector.
         * \param[in] lambda_IP     Vector of the lagrangian multipliers.
         *
         */
        virtual void update_Hessian(real const x[], real const lambda_IP[]) = 0;

      public:
        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                   protected Membervariablen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//
        /// Dimension of the statevector.
        int_type N_Statevector;
        /// Dimension of the constraints.
        int_type N_Constraintvector;
        /// Statevector.
        real * x;
        /// Vector of the lagrangian multipliers.
        real * lambda;
        /// Equidistant stepsize.
        real h;

        // Matrices of the hessian:
        /// Array of the matrices \f$W_i\f$.
        std::vector<MatrixWrapper<real>> Array_W;
        /// Array of the matrices \f$P_i\f$.
        std::vector<MatrixWrapper<real>> Array_P;
        /// Matrix \f$\phi\f$.
        MatrixWrapper<real> PHI;
        /// Matrix \f$P\f$.
        MatrixWrapper<real> MP;

        // Matrices of the jacobian:

        // ODE Constraints:
        /// Array of the matrices \f$M_i\f$.
        std::vector<MatrixWrapper<real>> Array_M;
        /// Array of the matrices \f$S_i\f$.
        std::vector<MatrixWrapper<real>> Array_S;
        /// Array of the matrices \f$R_i\f$.
        std::vector<MatrixWrapper<real>> Array_R;
        // Boundary Constraints:
        /// Upper matrix of the original boundary constraints.
        MatrixWrapper<real> b_Upper;
        /// Lower matrix of the original boundary constraints.
        MatrixWrapper<real> b_Lower;
        /// Matrix \f$R\f$.
        MatrixWrapper<real> R;
        // Nonlinear Constraints:
        /// Array of the matrices \f$G_i\f$.
        std::vector<MatrixWrapper<real>> Array_G;
        /// Array of the original matrices of parameters \f$Q_i\f$.
        std::vector<MatrixWrapper<real>> Array_q;

        // Corrected matrices for the interior-point-method:
        // Nonlinear Constraints:
        /// Array of the matrices C_CONSTR.
        std::vector<MatrixWrapper<real>> Array_C_CONSTR;
        /// Array of the matrices of parameters \f$Q_i\f$ in the interior-point-method.
        std::vector<MatrixWrapper<real>> Array_Q;
        /// Matrix \f$\hat{R}\f$.
        MatrixWrapper<real> R_Hat;
        // Parameter Constraints:
        /// Matrix \f$Q\f$.
        MatrixWrapper<real> Q;
        /// Upper matrix of the boundary constraints in the interior-point-method.
        MatrixWrapper<real> PSI_Upper;
        /// Lower matrix of the boundary constraints in the interior-point-method.
        MatrixWrapper<real> PSI_Lower;

        // Timestamps:
        /// Time for calculating the hessian.
        real TimeHesse;
        /// Time for calculating the jacobain.
        real TimeJacobi;
        /// Time for calculating the interior-point-method matrix structure.
        real TimeIP;
        /// Timer for getting the timestamps.
        TicToc MyTimer;
    };

} // namespace OCPbasic

#endif // DISKRETISIERUNG_H
