/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file trapez.h
 * Header definition of the class Trapez.
 *
 * @author A. Huber
 * @since 27.09.2018
 */

#ifndef TRAPEZ_H
#define TRAPEZ_H

#include "../linalg/OCPbasic_SymEigenValues.h"
#include "Discretization.h"

namespace OCPbasic
{

    template <typename real>
    class Trapez : public Discretization<real>
    {
      public:
        typedef MatrixWrapper<real> MatW;

        /**
         * \brief Trapez:
         *        Konstruktor der Klasse Trapez.
         *
         * \param[in] N     Anzahl der Zeitgitterpunkte.
         * \param[in] t0    Anfangszeitpunkt.
         * \param[in] tF    Endzeitpunkt.
         * \param[in] P     Optimalsteuerungsproblem.
         *
         */
        Trapez(int_type N, real t0, real tF, Problem<real> * P);

        /**
         * \brief ~Trapez:
         *        Virtueller Destruktor der Klasse Trapez.
         */
        virtual ~Trapez();

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                     public Memberfunktionen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief regularizeHessian:
         *        Regularize the hessian matrix.
         */
        virtual void regularizeHessian();

        /**
         * \brief writeToSolution:
         *        Writes the current data into a solution.
         *
         * \param[out] Sol      Solution object.
         *
         * @return True if the dimensions of the solution are correct.
         *
         */
        virtual bool writeToSolution(Solution<real> & Sol);

        /**
         * \brief readFromSolution:
         *        Reads the data from a solution.
         *
         * \param[in] Sol       Solution object.
         *
         * @return True if the dimensions of the solution are correct.
         *
         */
        virtual bool readFromSolution(Solution<real> & Sol);

        /**
         * \brief getP:
         *        Returns the pointer on the parameter array inside the statevector.
         *
         * @return Returns the pointer on the parameter array.
         *
         */
        virtual real * getP();

        /**
         * \brief getYk:
         *        Returns the pointer on the state \f$y_k\f$ inside the statevector.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the state \f$y_k\f$.
         *
         */
        virtual real * getYk(int_type Index);

        /**
         * \brief getUk:
         *        Returns the pointer on the control \f$u_k\f$ inside the statevector.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the control \f$u_k\f$.
         *
         */
        virtual real * getUk(int_type Index);

        /**
         * \brief getLAMBDAk:
         *        Returns the pointer on the lagrange mutlipliers \f$\lambda_k\f$.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the lagrange mutlipliers \f$\lambda_k\f$.
         *
         */
        virtual real * getLAMBDAk(int_type Index);

        /**
         * \brief getSIGMA:
         *        Returns the pointer on the lagrange mutlipliers \f$\sigma\f$.
         *
         * @return Returns the pointer on the lagrange mutlipliers \f$\sigma\f$.
         *
         */
        virtual real * getSIGMA();

        /**
         * \brief getMUk:
         *        Returns the pointer on the lagrange mutlipliers \f$\mu_k\f$.
         *
         * \param[in] Index     Index \f$k\f$.
         *
         * @return Returns the pointer on the lagrange mutlipliers \f$\mu_k\f$.
         *
         */
        virtual real * getMUk(int_type Index);

        /**
         * \brief getVARPHI:
         *        Retruns the value of the target function for a statevector.
         *
         * \param[in] states        Statevector.
         *
         * @return Retruns the target function for a statevector.
         *
         */
        virtual real getVARPHI(real const states[]);

        /**
         * \brief Grad_und_VARPHI:
         *        Retruns the value of the target function and its gradient for a statevector.
         *
         * \param[in]  states   Statevector.
         * \param[out] VARPHI   Target function value.
         * \param[out] Grad     Gradient of the target function.
         *
         */
        virtual void Grad_und_VARPHI(real const states[], real VARPHI[], real Grad[]);

        /**
         * \brief CON_Diskr:
         *        Evaluates the constraints for a statevector.
         *
         * \param[in]  states   Statevector.
         * \param[out] GG       Constraintvector.
         *
         */
        virtual void CON_Diskr(real const states[], real GG[]);

        /**
         * \brief update_IPData:
         *        Update of the interior point data.
         *
         * \param[in] x             Statevector.
         * \param[in] lambda_IP     Vector of the lagrangian multipliers.
         *
         */
        virtual void update_IPData(real const x[], real const lambda_IP[]);

        /**
         * \brief getDimStateIP:
         *        Returns the Dimension of the statevector in the IP.
         *
         * @return Returns the Dimension of the statevector in the IP.
         *
         */
        virtual int_type getDimStateIP();

        /**
         * \brief getDimLambdaIP:
         *        Returns the Dimension of the vector of the lagrangian multipliers in the IP.
         *
         * @return Returns the Dimension of the vector of the lagrangian multipliers in the IP.
         *
         */
        virtual int_type getDimLambdaIP();

        /**
         * \brief update_Jacobian:
         *        Updates the jacobian for a statevector.
         *
         * \param[in] x             Statevector.
         *
         */
        virtual void update_Jacobian(real const x[]);

        /**
         * \brief update_Hessian:
         *        Updates the hessian for a statevector and a vector of the lagrangian multipliers.
         *
         * \param[in] x             Statevector.
         * \param[in] lambda_IP     Vector of the lagrangian multipliers.
         *
         */
        virtual void update_Hessian(real const x[], real const lambda_IP[]);

      private:
        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                    private Memberfunktionen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief get_iDimHesse:
         *        Returns the dimensionarray for the hessian.
         *
         * \param[out] iDimHesse    Dimensionarray for the hessian.
         *
         */
        void get_iDimHesse(int_type iDimHesse[]);

        /**
         * \brief get_iDimJacobi:
         *        Returns the dimensionarray for the colums of the jacobian.
         *
         * \param[out] iDimJacobi   Dimensionarray for the colums of the jacobian.
         *
         */
        void get_iDimJacobi(int_type iDimJacobi[]);

        /**
         * \brief get_iDimIP:
         *        Returns the dimensionarray of the interior-point-matrix.
         *
         * \param[in] iDimIP        Dimensionarray of the interior-point-matrix.
         *
         */
        void get_iDimIP(int_type iDimIP[]);

        /**
         * \brief ClearJacobi:
         *        Clears the jacobian.
         */
        void ClearJacobi();

        /**
         * \brief ClearHesse:
         *        Clears the hessian.
         */
        void ClearHesse();

        /**
         * \brief JacobiSetODEConstraints:
         *        Sets the ODE-constraints in the jacobian.
         */
        void JacobiSetODEConstraints();

        /**
         * \brief JacobiSetBoundConstraints:
         *        Sets the boundary constraints in the jacobian.
         */
        void JacobiSetBoundConstraints();

        /**
         * \brief JacobiSetNLConstraints:
         *        Sets the nonlinear constraints in the jacobian.
         */
        void JacobiSetNLConstraints();

        /**
         * \brief HesseSetInitialFinalState:
         *        Sets the matrices for start and end states in the hessian.
         */
        void HesseSetInitialFinalState();

        /**
         * \brief HesseSetGridPointSubmatrix:
         *        Sets the matrices for all gridpoints in the hessian.
         */
        void HesseSetGridPointSubmatrix();

        /**
         * \brief HesseSetSumParameterMatrix:
         *        Sets the matrix of the sum of parameter matrices.
         */
        void HesseSetSumParameterMatrix();

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                     private Membervariablen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        IPbasic::Malloc<real> mem_real;
        /// Auxiliary matrix of dimension Ny x Ny:
        MatW EvalHelpNyNy1;
        /// Auxiliary matrix of dimension Ny x Ny:
        MatW EvalHelpNyNy2;
        /// Auxiliary matrix of dimension Ny x Ny:
        MatW EvalHelpNyNy3;
        /// Auxiliary matrix of dimension Ny x Nu:
        MatW EvalHelpNyNu1;
        /// Auxiliary matrix of dimension Ny x Nu:
        MatW EvalHelpNyNu2;
        /// Auxiliary matrix of dimension Ny x Nu:
        MatW EvalHelpNyNu3;
        /// Auxiliary matrix of dimension Ny x Np:
        MatW EvalHelpNyNp1;
        /// Auxiliary matrix of dimension Ny x Np:
        MatW EvalHelpNyNp2;

        /// Auxiliary matrix of dimension Np x Np:
        MatW EvalHelpNpNp1;
        /// Auxiliary matrix of dimension Np x Ny:
        MatW EvalHelpNpNy1;
        /// Auxiliary matrix of dimension Np x Nu:
        MatW EvalHelpNpNu1;
        /// Auxiliary matrix of dimension Nu x Nu:
        MatW EvalHelpNuNu1;
        /// Auxiliary matrix of dimension Nu x Nu:
        MatW EvalHelpNuNu2;
        /// Auxiliary matrix of dimension Nu x Nu:
        MatW EvalHelpNuNu3;
        /// Auxiliary matrix of dimension Nu x Np:
        MatW EvalHelpNuNp1;
        /// Auxiliary matrix of dimension Nu x Np:
        MatW EvalHelpNuNp2;
        /// Auxiliary matrix of dimension Nu x Ny:
        MatW EvalHelpNuNy1;
        /// Auxiliary matrix of dimension Nu x Ny:
        MatW EvalHelpNuNy2;

        /// Auxiliary matrix of dimension 1 x Ny:
        MatW EvalHelp1Ny1;
        /// Auxiliary matrix of dimension 1 x Ny:
        MatW EvalHelp1Ny2;
        /// Auxiliary matrix of dimension 1 x Np:
        MatW EvalHelp1Np1;

        /// Auxiliary matrix of dimension N_NLCTR x Ny:
        MatW EvalHelp_N_NLCTR_Ny;
        /// Auxiliary matrix of dimension N_NLCTR x Nu:
        MatW EvalHelp_N_NLCTR_Nu;

        /// Auxiliary vector f_k:
        real * EvalHelp_f_k;
        /// Auxiliary vector f_kk:
        real * EvalHelp_f_kk;
        /// Auxiliary vector g:
        real * EvalHelp_g;
        /// Auxiliary vector psi:
        real * EvalHelp_psi;
        /// Auxiliary vector eigens:
        real * EvalHelp_eigens;

        /// Generator of eigenvalues:
        OCPbasic::SymEigenValues<real> SymEigen;
    };

} // namespace OCPbasic

#endif // TRAPEZ_H
