/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file InteriorPointMethod.h
 * Header definition of the class InteriorPointMethod.
 *
 * @author A. Huber
 * @since 07.11.2018
 */

#ifndef INNEREPUNKTE_H
#define INNEREPUNKTE_H

#include "Optimizer.h"

#include "../OCP_Params.h"
#include <IPbasic/IPbasic.h>
#include <IPbasic/IPbasic_Parameter.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_interface.h>

namespace OCPbasic
{

    template <typename real>
    class IP_NLP;

    template <typename real>
    class IP_Solver;

    template <typename real>
    class InteriorPointMethod : public Optimizer<real>
    {
        /// Indexarray for sorting.
        int_type * SortIndArray;

        /// Solver of the linear system.
        IPbasic::LinearSolver_interface<real> * MyLinSolver;
        /// NLP for the discreete interior-point-method.
        IP_NLP<real> * MyNLP;
        /// NLP Solver of IPbasic.
        IP_Solver<real> * MyNLP_Solver;

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                  private Memberfunktionen                                                           //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief initStartSolution:
         *        Initializes the Optimizer with a initial guess.
         *
         * \param[in] Sol   Initial guess.
         *
         * @return True is dimensions are correct.
         *
         */
        bool initStartSolution(Solution<real> & Sol);

        /**
         *  \brief setIndexMap:
         *         Sets the internal index map.
         */
        void setIndexMap();

        /// Timer of the linear solver.
        TicToc MyTimerLinearSolver;
        /// Timestamp of the linear solver.
        real TimeLinearSolver;

      public:
        /**
         * \brief InteriorPointMethod :
         *        Constructor of the class InteriorPointMethod.
         *
         * \param[in] P                     Optimal control problem.
         * \param[in] NGit                  Number of Gridpoints.
         * \param[in] t0                    Startpoint in time.
         * \param[in] tF                    Endpoint in time.
         * \param[in] MyLinearSolverFlag    Endpoint in time.
         * \param[in] UsedLineSearchMethod  Used LineSearchMethod.
         *
         */
        InteriorPointMethod(Problem<real> *           P,
                            int_type                  NGit,
                            real                      t0                   = 0.0,
                            real                      tF                   = 1.0,
                            LinearSolverFlag          MyLinearSolverFlag   = OCPbasic::LINSOL_LAPACK_BANDED_KKT,
                            IPbasic::LineSearchMethod UsedLineSearchMethod = IPbasic::LINESEARCH_ARMIJO_L1_PENALTY);

        /**
         * \brief ~InteriorPointMethod:
         *        Virtual destructor of the class InteriorPointMethod.
         */
        virtual ~InteriorPointMethod();

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                   public Memberfunktionen                                                           //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief optimize:
         *        Optimizes the optimal control problem.
         *
         * \param[in,out] Sol   Input: Initial guess. Output: Solution of the optimizer.
         *
         * @return Returns if optimizing was successful.
         *
         */
        virtual bool optimize(Solution<real> & Sol);

        /**
         * \brief getLastIterationDataPoint:
         *        Get the last IterationDataPoint.
         *
         * @return Last IterationDataPoint.
         *
         */
        IPbasic::IterationDataPoint<real> getLastIterationDataPoint() const;

        /**
         * \brief getIPparameter:
         *        Retruns the parameter of the interior-point-method.
         *
         * @return Retruns the parameter of the interior-point-method.
         *
         */
        IPbasic::IPparameter<real> * getIPparameter();

        /**
         * \brief getOCPParameter:
         *        Retruns the parameter of the interior-point-method and the OCP.
         *
         * @return Retruns the parameter of the interior-point-method and the OCP.
         *
         */
        OCPbasic::OCP_Params<real> * getOCPParameter();

        /**
         * \brief getIPtimings:
         *        Return the IPtimings class.
         *
         * @return Return the IPtimings class.
         *
         */
        IPbasic::IPtimings<real> * getIPtimings();

        /**
         * \brief getOCPTimings:
         *        Return the OCP_Timings class.
         *
         * @return Return the OCP_Timings class.
         *
         */
        OCPbasic::OCP_Timings<real> * getOCPTimings();

      private:
        /**
         * \brief PrintTiming:
         *        Prints the timings to the terminal.
         */
        void PrintTiming();

        /// Timer to messure the initialization.
        TicToc LocalInitTimer;
    };

} // namespace OCPbasic

#endif // INNEREPUNKTE_H
