/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file Optimizer.h
 * Header definition of the class Optimizer.
 *
 * @author A. Huber
 * @since 23.10.2018
 */

#ifndef OPTIMIERER_H
#define OPTIMIERER_H

#include "../Discretization/Discretization.h"
#include "../problem.h"
#include "../solution.h"
#include <IPbasic/IPbasic_Data.h>

namespace OCPbasic
{

    template <typename real>
    class Optimizer
    {
      protected:
        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                  protected Membervariablenen                                                        //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /// Used discretisation.
        Discretization<real> * discretization;
        /// Optimal control problem.
        Problem<real> * P;
        /// Startpoint in time.
        real t0;
        /// Endpoint in time.
        real tF;

        /**
         * \brief Optimizer:
         *        Contructor of the class Optimizer.
         *
         * \param[in] P         Optimal control problem.
         * \param[in] t0        Startpoint in time.
         * \param[in] tF        Endpoint in time.
         *
         */
        Optimizer(Problem<real> * P, real t0, real tF)
            : P(P)
            , t0(t0)
            , tF(tF)
        {
        }

      public:
        /**
         * \brief ~Optimizer:
         *        Virtual destructor of the class Optimizer.
         */
        virtual ~Optimizer()
        {
        }

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                     Public Memberfunktionen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief getDiscretization:
         *        Returns the used discretisation.
         *
         * @return Returns the used discretisation.
         *
         */
        Discretization<real> * getDiscretization()
        {
            return this->discretization;
        }

        /**
         * \brief getProblem:
         *        Returns the optimal control problem.
         *
         * @return Returns the optimal control problem.
         *
         */
        Problem<real> * getProblem()
        {
            return this->P;
        }

        /**
         * \brief optimize:
         *        Optimizes the optimal control problem.
         *
         * \param[in,out] Sol   Input: Initial guess. Output: Solution of the optimizer.
         *
         * @return Returns if optimizing was successful.
         *
         */
        virtual bool optimize(Solution<real> & Sol) = 0;

        /**
         * \brief getLastIterationDataPoint:
         *        Get the last IterationDataPoint.
         *
         * @return Last IterationDataPoint.
         *
         */
        virtual IPbasic::IterationDataPoint<real> getLastIterationDataPoint() const = 0;
    };

} // namespace OCPbasic

#endif // OPTIMIERER_H
