/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef BANDEDLU_H
#define BANDEDLU_H
#include "../OCPbasic_config.h"

#if defined(OCPBASIC_USE_ALGLIN)

#else

#include <IPbasic/Errors/IPbasic_Errors.h>
#include <IPbasic/IPbasic.h>

using IPbasic::Errors;
using IPbasic::int_type;
using IPbasic::Malloc;

#ifndef CHECK_BANDED_DIM
//#define CHECK_BANDED_DIM
#endif

template <typename real>
class BandedLU
{
    /// Real memoryspace.
    Malloc<real> mem_real;
    /// Int_type memoryspace.
    Malloc<int_type> mem_int;

  public:
    /**
     * \brief BandedLU:
     *        Constructor of the class BandedLU.
     *
     * \param[in] Dim_RHS          Dimension of the banded matrix and the right-hand-side.
     * \param[in] SizeBand      Size of the band.
     *
     */
    BandedLU(int_type Dim_RHS, int_type SizeBand);

    /**
     * \brief ~BandedLU:
     *        Virtual Destructor of the class BandedLU.
     */
    virtual ~BandedLU();

    /**
     * \brief factorize:
     *        Factorize the banded matrix.
     *
     * @return True if the banded matrix could be factorized.
     *
     */
    bool factorize();

    /**
     * \brief solve:
     *        Solve factorized banded matrix with one right-hand-side.
     *
     * \param[in,out] RHS       Matrix of right-hand-side.
     *
     * @return True if system could be solved.
     *
     */
    bool solve(real RHS[]);

    void solver(real RHS[], bool & ret);

    /**
     * \brief solve:
     *        Solve factorized banded matrix with more right-hand-sides.
     *
     * \param[in]     NRHS      Number of right-hand-sides.
     * \param[in,out] RHS       Matrix of right-hand-sides.
     *
     * @return True if system could be solved.
     *
     */
    bool solve(int_type NRHS, real RHS[]);

    /**
     * \brief clear:
     *        Clear the banded matrix.
     */
    void clear();

    /**
     * \brief push_Val_C:
     *        Push value in c notation.
     *
     * \param[in] i     Index of the row.
     * \param[in] j     Index of the col.
     * \param[in] val   Value.
     *
     */
    inline void push_Val_C(int_type i, int_type j, real val)
    {
#ifdef CHECK_BANDED_DIM
        // Check banded dimension:
        if (std::max(0, j - this->KU) > i || i > std::min(this->Dim_RHS - 1, j + this->KL))
        {
            Errors::set_Error("Loaded Index out of bounds!");
            return;
        }
#endif
        this->AB[j * this->LDAB + this->KL + this->KU + i - j] = val;
    }

    inline void regularize(real tol)
    {
        for (int_type i = 0; i < this->Dim_RHS; i++)
            this->AB[i * this->LDAB + this->KL + this->KU + i - i] += tol;
    }

  protected:
    /// True if the banded matrix is factorized.
    bool isFactorized;
    /// Dimension of the banded matrix and the right-hand-side.
    int_type Dim_RHS;
    /// Size of the band.
    int_type SizeBand;
    /// Pivoting array.
    int_type * IPIV;
    /// Value array of the banded matrix.
    real * AB;
    /// The leading dimension of the array AB. LDAB >= 2*KL+KU+1.
    int_type LDAB;
    /// Number of subdiagonals.
    int_type KL;
    /// Number of superdiagonals.
    int_type KU;

    /**
     * \brief gbtrf: (LAPACK)
     *        Computes an LU factorization of a real m-by-n band matrix A
     *        using partial pivoting with row interchanges.
     *
     * \param[in]     M         The number of rows of the matrix A.  M >= 0.
     * \param[in]     N         The number of columns of the matrix A.  N >= 0.
     * \param[in]     KL        The number of subdiagonals within the band of A.  KL >= 0.
     * \param[in]     KU        The number of superdiagonals within the band of A.  KU >= 0.
     * \param[in,out] AB        Value array of the banded matrix.
     * \param[in]     LDAB      The leading dimension of the array AB.  LDAB >= 2*KL+KU+1.
     * \param[out]    IPIV      The pivot indices; for 1 <= i <= min(M,N), row i of the
     *                          matrix was interchanged with row IPIV(i).
     *
     * @return = 0: successful exit
     *         < 0: if INFO = -i, the i-th argument had an illegal value
     *         > 0: if INFO = +i, U(i,i) is exactly zero. The factorization
     *              has been completed, but the factor U is exactly
     *              singular, and division by zero will occur if it is used
     *              to solve a system of equations.
     *
     */
    int_type gbtrf(int_type M, int_type N, int_type KL, int_type KU, real AB[], int_type LDAB, int_type IPIV[]);

    /**
     * \brief gbtrs: (LAPACK)
     *        Solves a system of linear equations
     *        A * X = B  or  A**T * X = B
     *        with a general band matrix A using the LU factorization computed
     *        by DGBTRF.
     *
     * \param[in]     TRANS     Specifies the form of the system of equations.
     *                          = 'N':  A * X = B  (No transpose)
     *                          = 'T':  A**T* X = B  (Transpose)
     *                          = 'C':  A**T* X = B  (Conjugate transpose = Transpose)
     * \param[in]     N         The order of the matrix A.  N >= 0.
     * \param[in]     KL        The number of subdiagonals within the band of A.  KL >= 0.
     * \param[in]     KU        The number of superdiagonals within the band of A.  KU >= 0.
     * \param[in]     NRHS      The number of right hand sides, i.e., the number of columns
                                of the matrix B.  Dim_RHS >= 0.
     * \param[in]     AB        Value array of the banded matrix.
     * \param[in]     LDAB      The leading dimension of the array AB.  LDAB >= 2*KL+KU+1.
     * \param[in]     IPIV      The pivot indices; for 1 <= i <= N, row i of the matrix was
                                interchanged with row IPIV(i).
     * \param[in,out] B         On entry, the right hand side matrix B.
                                On exit, the solution matrix X.
     * \param[in]     LDB       The leading dimension of the array B.  LDB >= max(1,N).
     *
     * @return = 0:  successful exit
     *         < 0: if INFO = -i, the i-th argument had an illegal value
     *
     */
    int_type gbtrs(char TRANS, int_type N, int_type KL, int_type KU, int_type NRHS, real AB[], int_type LDAB, int_type IPIV[], real B[], int_type LDB);
};

#endif

#endif // BANDEDLU_H
