/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef DENSELU_H
#define DENSELU_H

#include "../OCPbasic_config.h"

#if defined(OCPBASIC_USE_ALGLIN)

#else

#include <IPbasic/Errors/IPbasic_Errors.h>
#include <IPbasic/IPbasic.h>

using IPbasic::Errors;
using IPbasic::int_type;
using IPbasic::Malloc;

template <typename real>
class DenseLU
{
    /// Real memoryspace.
    Malloc<real> mem_real;
    /// Int_type memoryspace.
    Malloc<int_type> mem_int;

  public:
    /**
     * \brief DenseLU:
     *        Constructor of the class DenseLU.
     *
     * \param[in] DimRHS    Dimension of the dense matrix and the right-hand-side.
     *
     */
    DenseLU(int_type DimRHS);

    /**
     * \brief ~DenseLU:
     *        Virtual destructor of the class DenseLU.
     */
    virtual ~DenseLU();

    /**
     * \brief getValueArray:
     *        Get the value array of the dense matrix.
     *
     * @return Returns the value array of the dense matrix.
     */
    real * getValueArray();

    /**
     * \brief getDimension:
     *        Get the dimension of the dense matrix.
     *
     * @return Returns the dimension of the dense matrix.
     *
     */
    int_type getDimension();

    /**
     * \brief factorize:
     *        Factorize the dense matrix.
     *
     * @return True if the dense matrix could be factorized.
     *
     */
    bool factorize();

    /**
     * \brief solve:
     *        Solve factorized dense matrix with one right-hand-side.
     *
     * \param[in,out] RHS       Matrix of right-hand-side.
     *
     * @return True if system could be solved.
     *
     */
    bool solve(real RHS[]);

    /**
     * \brief solve:
     *        Solve factorized dense matrix with more right-hand-sides.
     *
     * \param[in]     NRHS      Number of right-hand-sides.
     * \param[in,out] RHS       Matrix of right-hand-sides.
     *
     * @return True if system could be solved.
     *
     */
    bool solve(int_type NRHS, real RHS[]);

    /**
     * \brief clear:
     *        Clear the dense matrix.
     */
    void clear();

    /**
     * \brief push_Val_C:
     *        Push value in c notation.
     *
     * \param[in] i     Index of the row.
     * \param[in] j     Index of the col.
     * \param[in] val   Value.
     *
     */
    inline void push_Val_C(int_type i, int_type j, real val)
    {
        this->A[j * this->LDA + i] = val;
    }

  protected:
    /// Dimension of the dense matrix and the right-hand-side.
    int_type DimRHS;
    /// Array of the dense matrix.
    real * A;
    /// The leading dimension of the array A.  LDA >= max(1,M).
    int_type LDA;
    /// The pivot indices; for 1 <= i <= min(M,N), row i of the
    /// matrix was interchanged with row IPIV(i).
    int_type * IPIV;

    /**
     * \brief getrf: (LAPACK)
     *        Factorize the matrix.
     *
     * \param[in]     M         The number of rows of the matrix A.  M >= 0.
     * \param[in]     N         The number of columns of the matrix A.  N >= 0.
     * \param[in,out] A         Array of the dense matrix.
     * \param[in]     LDA       The leading dimension of the array A.  LDA >= max(1,M).
     * \param[out]    IPIV      The pivot indices; for 1 <= i <= min(M,N), row i of the
     *                      matrix was interchanged with row IPIV(i).
     *
     * @return = 0:  successful exit
     *         < 0:  if INFO = -i, the i-th argument had an illegal value
     *         > 0:  if INFO = i, U(i,i) is exactly zero. The factorization
     *               has been completed, but the factor U is exactly
     *               singular, and division by zero will occur if it is used
     *               to solve a system of equations.
     *
     */
    int_type getrf(int_type M, int_type N, real A[], int_type LDA, int_type IPIV[]);

    /**
     * \brief getrs: (LAPACK)
     *        Solve the factorized dense system with right-hand-sides.
     *
     * \param[in]     TRANS     Specifies the form of the system of equations:
     *                          = 'N':  A * X = B  (No transpose)
     *                          = 'T':  A**T* X = B  (Transpose)
     *                          = 'C':  A**T* X = B  (Conjugate transpose = Transpose)
     * \param[in]     N         The order of the matrix A.  N >= 0.
     * \param[in]     NRHS      The number of right hand sides, i.e., the number of columns
     *                          of the matrix B.  NRHS >= 0.
     * \param[in]     A         Array of the dense matrix.
     * \param[in]     LDA       The leading dimension of the array A.  LDA >= max(1,N).
     * \param[in]     IPIV      The pivot indices from DGETRF; for 1<=i<=N, row i of the
     *                          matrix was interchanged with row IPIV(i).
     * \param[in,out] B         On entry, the right hand side matrix B.
     *                          On exit, the solution matrix X.
     * \param[in]     LDB       The leading dimension of the array B.  LDB >= max(1,N).
     *
     * @return = 0:  successful exit
     *         < 0:  if INFO = -i, the i-th argument had an illegal value
     *
     */
    int_type getrs(char TRANS, int_type N, int_type NRHS, real A[], int_type LDA, int_type IPIV[], real B[], int_type LDB);
};

#endif

#endif // DENSELU_H
