/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef LAPACKKKTSOLVER_H
#define LAPACKKKTSOLVER_H

#include <IPbasic/Errors/IPbasic_Errors.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_interface.h>

#include "../OCPbasic_config.h"

#if defined(OCPBASIC_USE_ALGLIN)
#include <Alglin/Alglin++.hh>
#include <Alglin/KKT_like.hh>

using IPbasic::Errors;
using IPbasic::int_type;
using IPbasic::Malloc;

template <typename real>
class LapackKKTSolver : public IPbasic::LinearSolver_interface<real>
{
    /// Int_type memoryspace.
    Malloc<int_type> mem_int;

  public:
    /**
     * \brief LapackKKTSolver:
     *        Constructor of the class LapackKKTSolver.
     *
     * \param[in] DimRHS            Dimension of right-hand-side.
     * \param[in] NDimRowBand       Dimension of rows of the band.
     * \param[in] SizeBand          Size of the band.
     * \param[in] SortIndexArray    Sorting array of sparsematrix.
     *
     */
    LapackKKTSolver(int_type DimRHS, int_type NDimRowBand, int_type SizeBand, int_type * SortIndexArray);

    /**
     * \brief ~LapackKKTSolver:
     *        Virtual destructor of the class LapackKKTSolver.
     */
    virtual ~LapackKKTSolver();

    /*!
     * \brief factorize:
     *        Factorizes the matrix \f$ A\f$ for a sparse matrix.
     *
     * \param[in] Sparse               Sparse matrix which has to be solved.
     * \param[in] is_stored_symmetric  If the flag is `true` only the lower part is stored.
     *
     * \return True if the the matrix \f$ A\f$ could be factorized.
     *
     */
    virtual bool factorize(IPbasic::SparseMatrixBase<real> const & Sparse, bool is_stored_symmetric);

    /*!
     * \brief solve:
     *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$
     *        for a sparse matrix.
     *
     * \param[in,out] RHS     In:  right-hand-side of the linear system.
     *                        OUT: solution of the linear system.
     *
     * \return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
     *
     */
    virtual bool solve(real RHS[]);

    /**
     * \brief getTimes:
     *        Get Solver Timings.
     *
     * \param[out] TimeFactorizeKKT      Time to factorize the KKT matrix.
     * \param[out] TimeSolveKKT         Time to solve the KKT matrix.
     * \param[out] TimeFactorizeBanded   Time to factorize the banded matrix.
     * \param[out] TimeSolveBanded      Time to solve the banded matrix.
     * \param[out] TimeFactorizeDense   Time to factorize the dense matrix.
     * \param[out] TimeSolveDense       Time to solve the dense matrix.
     *
     */
    void getTimes(real & TimeFactorizeKKT, real & TimeSolveKKT, real & TimeFactorizeBanded, real & TimeSolveBanded, real & TimeFactorizeDense, real & TimeSolveDense);

  private:
    /// Dimension of right-hand-side.
    int_type DimRHS;
    /// Dimension of rows of the band.
    int_type NDimRowBand;
    /// Size of the band.
    int_type SizeBand;
    /// Dimension of rows of the rest matrix.
    int_type NRest;
    /// Sorting array of sparsematrix.
    int_type * SortIndexArray;
    /// Temporary copy array.
    real * copyarray;
    /// Bandmatrix KKT-Solver.
    alglin::KKT<real> * AlglinSolver;
    /// Time to factorize the KKT matrix.
    real TimeFactorizeKKT;
    /// Time to solve the KKT matrix.
    real TimeSolveKKT;
    /// Time to factorize the banded matrix.
    real TimeFactorizeBanded;
    /// Time to solve the banded matrix.
    real TimeSolveBanded;
    /// Time to factorize the dense matrix.
    real TimeFactorizeDense;
    /// Time to solve the dense matrix.
    real TimeSolveDense;
};

#else

using IPbasic::Errors;
using IPbasic::int_type;
using IPbasic::Malloc;
using IPbasic::MatrixWrapper;

// Dummy definition of the class BandedLU.
template <typename real>
class BandedLU;
// Dummy definition of the class DenseLU.
template <typename real>
class DenseLU;

template <typename real>
class LapackKKTSolver : public IPbasic::LinearSolver_interface<real>
{
    /// Real memoryspace.
    Malloc<real> mem_real;
    /// Int_type memoryspace.
    Malloc<int_type> mem_int;

  public:
    /**
     * \brief LapackKKTSolver:
     *        Constructor of the class LapackKKTSolver.
     *
     * \param[in] DimRHS            Dimension of right-hand-side.
     * \param[in] NDimRowBand       Dimension of rows of the band.
     * \param[in] SizeBand          Size of the band.
     * \param[in] SortIndexArray    Sorting array of sparsematrix.
     *
     */
    LapackKKTSolver(int_type DimRHS, int_type NDimRowBand, int_type SizeBand, int_type * SortIndexArray);

    /**
     * \brief ~LapackKKTSolver:
     *        Virtual destructor of the class LapackKKTSolver.
     */
    virtual ~LapackKKTSolver();

    /*!
     * \brief factorize:
     *        Factorizes the matrix \f$ A\f$ for a sparse matrix.
     *
     * \param[in] Sparse               Sparse matrix which has to be solved.
     * \param[in] is_stored_symmetric  If the flag is `true` only the lower part is stored.
     *
     * \return True if the the matrix \f$ A\f$ could be factorized.
     *
     */
    virtual bool factorize(IPbasic::SparseMatrixBase<real> const & Sparse, bool is_stored_symmetric);

    /*!
     * \brief solve:
     *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$
     *        for a sparse matrix.
     *
     * \param[in,out] RHS     In:  right-hand-side of the linear system.
     *                        OUT: solution of the linear system.
     *
     * \return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
     *
     */
    virtual bool solve(real RHS[]);

    /**
     * \brief getTimes:
     *        Get Solver Timings.
     *
     * \param[out] TimeFactorizeKKT      Time to factorize the KKT matrix.
     * \param[out] TimeSolveKKT         Time to solve the KKT matrix.
     * \param[out] TimeFactorizeBanded   Time to factorize the banded matrix.
     * \param[out] TimeSolveBanded      Time to solve the banded matrix.
     * \param[out] TimeFactorizeDense   Time to factorize the dense matrix.
     * \param[out] TimeSolveDense       Time to solve the dense matrix.
     *
     */
    void getTimes(real & TimeFactorizeKKT, real & TimeSolveKKT, real & TimeFactorizeBanded, real & TimeSolveBanded, real & TimeFactorizeDense, real & TimeSolveDense);

  private:
    /// Dimension of right-hand-side.
    int_type DimRHS;
    /// Dimension of rows of the band.
    int_type NDimRowBand;
    /// Size of the band.
    int_type SizeBand;
    /// Dimension of rows of the rest matrix.
    int_type NRest;
    /// Sorting array of sparsematrix.
    int_type * SortIndexArray;
    /// Temporary copy array.
    real * copyarray;
    /// Banded LU solver.
    BandedLU<real> * MyBandedLUSolver;
    /// Dense LU solver.
    DenseLU<real> * MyDenseLUSolver;
    /// Memory of the submatrix \f$A^top\f$.
    real * Mem_A_top;
    /// Memory of the submatrix \f$A^top_2\f$.
    real * Mem_A_top2;
    /// Memory of B.
    real * Mem_B;
    /// Submatrix \f$A^top\f$.
    MatrixWrapper<real> A_top;
    /// Submatrix \f$A^top_2\f$.
    MatrixWrapper<real> A_top2;
    /// Submatrix \f$B\f$.
    MatrixWrapper<real> B;
    /// Helperarray real_hlp.
    real * real_hlp;
    /// Time to factorize the KKT matrix.
    real TimeFactorizeKKT;
    /// Time to solve the KKT matrix.
    real TimeSolveKKT;
    /// Time to factorize the banded matrix.
    real TimeFactorizeBanded;
    /// Time to solve the banded matrix.
    real TimeSolveBanded;
    /// Time to factorize the dense matrix.
    real TimeFactorizeDense;
    /// Time to solve the dense matrix.
    real TimeSolveDense;
    /// /// Timer to messure KKT times.
    TicToc KKT_Timer;
    /// Timer to messure internal times.
    TicToc Internal_Timer;

    /**
     * \brief loadSymSparse_C:
     *        Load a symmetric sparse matrix.
     *
     * \param[in] nnz       Number of non zeros.
     * \param[in] rows      Row array of the sparse matrix.
     * \param[in] cols      Column array of the sparse matrix.
     * \param[in] vals      Value array of the sparse matrix.
     *
     */
    void loadSymSparse_C(int_type nnz, const int_type rows[], const int_type cols[], const real vals[]);

    /**
     * \brief loadSymSparse_Fortran:
     *        Load a symmetric sparse matrix.
     *
     * \param[in] nnz       Number of non zeros.
     * \param[in] rows      Row array of the sparse matrix.
     * \param[in] cols      Column array of the sparse matrix.
     * \param[in] vals      Value array of the sparse matrix.
     *
     */
    void loadSymSparse_Fortran(int_type nnz, const int_type rows[], const int_type cols[], const real vals[]);

    /**
     * \brief clear:
     *        Clear the matrix memory.
     */
    void clear();

    /**
     * \brief gemv:
     *        Matrix vector multipication. (alpha * OP(A) * x + beta * y)
     *
     * \param[in]     transA    Flag for transposed A.
     * \param[in]     M         Number of rows.
     * \param[in]     N         Number of columns.
     * \param[in]     alpha     Scalar factor.
     * \param[in]     A         Value array of A.
     * \param[in]     LDA       Leading dimension of A.
     * \param[in]     X         Vector array of x.
     * \param[in]     incX      Increment of x.
     * \param[in]     beta      Scalar factor.
     * \param[in,out] Y         Vector array of y.
     * \param[in]     incY      Increment of y.
     *
     */
    void gemv(bool transA, int_type M, int_type N, real alpha, real A[], int_type LDA, real X[], int_type incX, real beta, real Y[], int_type incY);

    /**
     * \brief gemm:
     *        Matrix matrix multipication. (alpha * OP(A) * OP(B) + beta * C)
     *
     * \param[in]     transA    Flag for transposed A.
     * \param[in]     transB    Flag for transposed B.
     * \param[in]     M         Specifies  the number  of rows  of the  matrix
     *                          op( A )  and of the  matrix  C.  M  must  be at least  zero.
     * \param[in]     N         Specifies the number  of columns of the matrix
     *                          op( B ) and the number of columns of the matrix C. N must be
     *                          at least zero.
     * \param[in]     K         Apecifies  the number of columns of the matrix
     *                          op( A ) and the number of rows of the matrix op( B ). K must
     *                          be at least  zero.
     * \param[in]     alpha     Specifies the scalar alpha.
     * \param[in]     A         Value array of A.
     * \param[in]     LDA       Leading dimension of A.
     * \param[in]     B         Value array of B.
     * \param[in]     LDB       Leading dimension of B.
     * \param[in]     beta      Specifies the scalar  beta.
     * \param[in,out] C         Value array of C.
     * \param[in]     LDC       Leading dimension of C.
     *
     */
    void gemm(bool transA, bool transB, int_type M, int_type N, int_type K, real alpha, real A[], int_type LDA, real B[], int_type LDB, real beta, real C[], int_type LDC);
};

#endif

#endif // LAPACKKKTSOLVER_H
