/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef OCPBASIC_HOUSEHOLDERQR_H
#define OCPBASIC_HOUSEHOLDERQR_H

#include <IPbasic/IPbasic.h>

using IPbasic::int_type;
using IPbasic::Malloc;
using IPbasic::MatrixWrapper;

namespace OCPbasic
{
    /**
     * \brief enorm:
     *        Computes the euclidean norm over a vector \f$\|\vec{a}\|_2\f$.
     *
     * \param[in] N         Dimension of the vector.
     * \param[in] vec       Given vector.
     *
     * @return The euclidean norm over a vector \f$\|\vec{a}\|_2\f$.
     *
     */
    template <typename real>
    static real enorm(int_type N, const real vec[])
    {
        real sum = real(0);
        for (int_type i = 0; i < N; ++i)
            sum += vec[i] * vec[i];
        return std::sqrt(sum);
    }

    /**
     * \brief dot:
     *        Computes the scalarproduct of two vectors \f$\vec{a} \cdot \vec{b}\f$.
     *
     * \param[in] N     Dimension of the vectors.
     * \param[in] a     Given vector \f$\vec{a}\f$.
     * \param[in] b     Given vector \f$\vec{b}\f$.
     *
     * @return The scalarproduct of two vectors \f$\vec{a} \cdot \vec{b}\f$.
     *
     */
    template <typename real>
    static real dot(int_type N, real const a[], real const b[])
    {
        real sum = real(0.0);
        for (int_type i = 0; i < N; ++i)
            sum += a[i] * b[i];
        return sum;
    }

    /**
     * \brief QR:
     *        QR decomposition of an matrix \f$M \in \mathbb{R}^{n\times m}\f$.
     *
     * \param[in,out] M                 Matrix.
     * \param[out]    QR_aux            Auxiliary array with missing diagonal values of vi. Minimum length = min(n_row, n_col)
     * \param[out]    ipvt              Pivotarray of M.
     * \param[in]     tol               Tolerance used in Rank Estimaltion.
     * \param[in]     Flag_Pivoting     Flag if using pivot strategy.
     * \param[out]    rank              Rank of M (Only if Flag_Pivoting == true).
     *
     * @return Error flag.
     *
     */
    template <typename real>
    static int_type QR(MatrixWrapper<real> & M, int_type ipvt[], real QR_aux[], real tol, bool Flag_Pivoting, int_type & rank)
    {
        int_type nr  = M.numRows();
        int_type nc  = M.numCols();
        int_type lda = M.lDim();
        if (nr < 0 || nc < 0)
            return -1;
        if (tol < 0)
            return -1;
        if (lda < nr)
            return -1;

        // Allocate workingspace:
        Malloc<real> mem("QR");
        mem.allocate(size_t(nr + 2 * nc));
        real * v_i  = mem(size_t(nr));
        real * w_i  = mem(size_t(nc));
        real * Norm = mem(size_t(nc));

        // Set Data:
        for (int_type i = 0; i < std::min(nr, nc); ++i)
            QR_aux[i] = real(0.0);

        rank             = 0;
        int_type jmax    = 0;
        real     tau     = real(0.0);
        real     sigma   = real(0.0);
        real     Aii     = real(0.0);
        real     V1      = real(0.0);
        real     beta    = real(0.0);
        real     sum     = real(0.0);
        real     signAii = real(1.0);

        // Compute Norm of columns:
        for (int_type j = 0; j < nc; ++j)
        {
            Norm[j] = enorm(nr, &M(0, j));
            if (Norm[j] > tau)
            {
                tau  = Norm[j];
                jmax = j;
            }
            ipvt[j] = j;
        }
        // Check rank:
        if (IPbasic::isZero(tau))
        {
            rank = 0;
            return -2;
        }

        // Householder loop:
        for (int_type i = 0; i < std::min(nr - 1, nc); ++i)
        {
            if (Flag_Pivoting)
            {
                // Swap pivot:
                std::swap(ipvt[i], ipvt[jmax]);
            }
            // Perform householder transformation:
            sigma = Norm[ipvt[i]];
            if (IPbasic::isZero(sigma))
            {
                QR_aux[i] = 0;
            }
            else
            {
                Aii     = M(i, ipvt[i]);
                signAii = real(Aii <= 0 ? -1 : 1);
                V1      = (Aii / sigma) + signAii;
                // Compute v_i:
                v_i[i] = V1;
                for (int_type k = i + 1; k < nr; ++k)
                    v_i[k] = M(k, ipvt[i]) / sigma;

                // Compute beta:
                beta      = real(1.0) / (real(1.0) + std::abs(Aii) / sigma);
                QR_aux[i] = beta;
                // Set workingarray for size of columns: w = beta*A'*V
                for (int_type k = i; k < nc; ++k)
                    w_i[k] = beta * dot(nr - i, &M(i, ipvt[k]), &v_i[i]);
                // Compute Housholder*A = A - V*W'
                for (int_type k = i; k < nc; ++k)
                    for (int_type q = i; q < nr; ++q)
                        M(q, ipvt[k]) = M(q, ipvt[k]) - (v_i[q] * w_i[k]);
                // Store v_i in A:
                for (int_type k = i + 1; k < nr; ++k)
                    M(k, ipvt[i]) = v_i[k];
            }
            // Adjust norm of columns:
            jmax = 0;
            tau  = real(0);
            for (int_type k = i + 1; k < nc; ++k)
            {
                // Update of euler norm:
                sum = std::sqrt(Norm[ipvt[k]] * Norm[ipvt[k]] - M(i, ipvt[k]) * M(i, ipvt[k]));
                if (IPbasic::isNaN(sum))
                    sum = real(0); // sum = this->enorm(nr-k, &M(i+1, ipvt[k]));

                Norm[ipvt[k]] = sum;
                if (sum > tau)
                {
                    tau  = sum;
                    jmax = k;
                }
            }

            // Exit if tau == 0
            if (IPbasic::isZero(tau))
                break;
        }
        // Estimate rank of A:
        if (Flag_Pivoting)
        {
            rank = 0;
            for (int_type i = 0; i < std::min(nr, nc); ++i)
            {
                if (std::abs(M(i, ipvt[i])) > tol * std::abs(M(1, ipvt[1])))
                    rank = i + 1;
            }
        }
        else
        {
            rank = -1;
        }
        return 0;
    }

    template <typename real>
    int_type QMult_Right(MatrixWrapper<real> const & QR, int_type ipvt[], real QR_aux[], MatrixWrapper<real> & B)
    {
        // Check Dimensions:
        int_type nr  = QR.numRows();
        int_type nc  = QR.numCols();
        int_type Br  = B.numRows();
        int_type lda = QR.lDim();
        int_type ldb = B.lDim();

        if (nr < 0 || nc < 0 || Br < 0)
            return -1;
        if (lda < nr || ldb < Br)
            return -1;
        if (nr != B.numCols())
            return -1;

        // Allocate Workspace:
        Malloc<real> mem("QMult_Right");
        mem.allocate(size_t(nr + Br));
        real * v_i = mem(size_t(nr));
        real * w_i = mem(size_t(Br));

        real beta = real(0.0);

        // loop:
        for (int_type i = 0; i < std::min(nr - 1, nc); ++i)
        {
            // Assing v_i:
            for (int_type k = i + 1; k < nr; ++k)
                v_i[k] = QR(k, ipvt[i]);
            beta = QR_aux[i];
            if (IPbasic::isZero(beta))
            {
                v_i[i] = real(0.0);
            }
            else
            {
                v_i[i] = real(1.0) / beta;
                if (QR(i, ipvt[i]) > real(0.0))
                    v_i[i] *= real(-1.0);
            }
            // Compute W = beta*B*V
            for (int_type j = 0; j < Br; ++j)
            {
                w_i[j] = real(0.0);
                for (int_type k = i; k < nr; ++k)
                    w_i[j] += B(j, k) * v_i[k];
                w_i[j] *= beta;
            }
            // Compute B*H = B - W*V'
            for (int_type j = 0; j < Br; ++j)
                for (int_type k = i; k < nr; ++k)
                    B(j, k) -= v_i[k] * w_i[j];
        }
        return 0;
    }

    /**
     * \brief get_R:
     *        Get R form householderQR matrix and data.
     *
     * \param[in]  QR           HouseholderQR matrix
     * \param[in]  ipvt         Pivoting array.
     * \param[out] R            R matrix of QR.
     *
     * @return 0, if dimensions are ok.
     *
     */
    template <typename real>
    static int_type get_R(MatrixWrapper<real> const & QR, int_type ipvt[], MatrixWrapper<real> & R)
    {
        // Check Dimensions:
        int_type nr  = QR.numRows();
        int_type nc  = QR.numCols();
        int_type lda = QR.lDim();
        if (nr < 0 || nc < 0)
            return -1;
        if (lda < nr)
            return -1;
        if (nr != R.numRows() || nc != R.numCols() || lda != R.lDim())
            return -1;
        // Get R:
        R.fill(real(0.0));
        for (int_type i = 0; i < nr; ++i)
            for (int_type j = 0; j < nc; ++j)
                if (j >= i)
                    R(i, j) = QR(i, ipvt[j]);
        return 0;
    }

    /**
     * \brief get_Q:
     *        Get Q form householderQR matrix and data.
     *
     * \param[in]     QR           HouseholderQR matrix
     * \param[in]     ipvt         Pivoting array.
     * \param[in]     QR_aux       HouseholderQR auxilary array (\f$\beta\f$).
     * \param[out]    Q            Q matrix of QR.
     *
     * @return 0, if dimensions are ok.
     *
     */
    template <typename real>
    static int_type get_Q(MatrixWrapper<real> const & QR, int_type ipvt[], real QR_aux[], MatrixWrapper<real> & Q)
    {
        // Check Dimensions:
        int_type nr  = QR.numRows();
        int_type nc  = QR.numCols();
        int_type lda = QR.lDim();
        if (nr < 0 || nc < 0)
            return -1;
        if (lda < nr)
            return -1;
        if (nr != Q.numRows() || nr != Q.numCols() || lda != Q.lDim())
            return -1;
        // Get temporary memory:
        Malloc<real> mem("get_Q");
        mem.allocate(size_t(2 * nr));
        real * v_i = mem(size_t(nr));
        real * w_i = mem(size_t(nr));

        real beta = real(0.0);
        // Initialize Q with I:
        Q.fill(real(0.0));
        for (int_type i = 0; i < nr; ++i)
            Q(i, i) = real(1.0);
        // Reverse loop over v_i:
        for (int_type i = std::min(nr - 1, nc) - 1; i >= 0; --i)
        {
            // Assign v_i:
            for (int_type k = i + 1; k < nr; ++k)
                v_i[k] = QR(k, ipvt[i]);
            beta = QR_aux[i];
            if (IPbasic::isZero(beta))
            {
                v_i[i] = real(0.0);
            }
            else
            {
                v_i[i] = real(1.0) / beta;
                if (QR(i, ipvt[i]) > 0)
                    v_i[i] *= real(-1);
            }
            // Compute w_i = beta*Q'*v_i
            for (int_type k = i; k < nr; ++k)
                w_i[k] = beta * dot(nr - i, &Q(i, k), &v_i[i]);
            // Compute H*Q = Q - v*w'
            for (int_type k = i; k < nr; ++k)
                for (int_type l = i; l < nr; ++l)
                    Q(l, k) -= v_i[l] * w_i[k];
        }
        return 0;
    }

    /**
     * \brief get_Q_and_R:
     *        Get Q and R form householderQR matrix and data.
     *
     * \param[in]     QR           HouseholderQR matrix
     * \param[in]     ipvt         Pivoting array.
     * \param[in]     QR_aux       HouseholderQR auxilary array (\f$\beta\f$).
     * \param[out]    Q            Q matrix of QR.
     * \param[out]    R            R matrix of QR.
     *
     * @return 0, if dimensions are ok.
     *
     */
    template <typename real>
    static int_type get_Q_and_R(MatrixWrapper<real> const & QR, int_type ipvt[], real QR_aux[], MatrixWrapper<real> & Q, MatrixWrapper<real> & R)
    {
        int_type ret = get_R(QR, ipvt, R);
        if (0 == ret)
            ret = get_Q(QR, ipvt, QR_aux, Q);
        return ret;
    }
} // namespace OCPbasic

#endif // OCPBASIC_HOUSEHOLDERQR_H
