/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file collectormatrix.h
 * Header Definition der Klasse CollectorMatrix.
 *
 * @author A. Huber
 * @since 08.12.2016
 */

#ifndef COLLECTORMATRIX_H
#define COLLECTORMATRIX_H

#include "sparsematrix.h"

template <typename real>
class CollectorMatrix
{
  public:
    /**
     * \brief CollectorMatrix:
     *        Konstruktor der Klasse CollectorMatrix.
     *
     * \param[in] N_Coll_Row  Anzahl der Collector Rows.
     * \param[in] N_Coll_Col  Anzahl der Collector Columns.
     * \param[in] dimRow      DimensionenArray der Collector Rows.
     * \param[in] dimCol      DimensionenArray der Collector Columns.
     * \param[in] Flag_Sym    Flag, der angibt, ob Daten in der CollectorMatrix symmetrisch
     *                        abgelegt werden.
     *
     */
    CollectorMatrix(int_type N_Coll_Row, int_type N_Coll_Col, int_type const dimRow[], int_type const dimCol[], isSym Flag_Sym = NOT_SYM);

    /**
     * \brief ~CollectorMatrix:
     *        Virtueller Destruktor der Klasse CollectorMatrix.
     */
    virtual ~CollectorMatrix();

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                     public Memberfunktionen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief getFlag_Sym:
     *        Gibt den aktuellen Symmetrieflag zurück.
     *
     * @return Aktueller Symmetrieflag.
     *
     */
    isSym getFlag_Sym();

    /**
     * \brief getNMaxSparse:
     *        Gibt die aktuelle Anzahl der Einträge der resultierenden Sparsematrix zurück.
     *
     * @return Gibt die aktuelle Anzahl der Einträge der resultierenden Sparsematrix zurück.
     */
    int_type getNMaxSparse();

    /**
     * \brief getSparseMatrix:
     *        Gibt die resultierende Sparsematrix zurück.
     *
     * \param[out] Sparse   Resultierende Sparsematrix.
     *
     */
    void getSparseMatrix(SparseMatrix<real> & Sparse);

    /**
     * \brief getMatrix:
     *        Gibt die resultierende Matrix zurück.
     *
     * \param[out] M        Resultierende Matrix.
     *
     */
    void getMatrix(MatrixWrapper<real> & M);

    /**
     * \brief getRowDimSparse:
     *        Gibt die row dimension der resultierenden SparseMatrix zurück.
     *
     * @return Gibt die row dimension der resultierenden SparseMatrix zurück.
     *
     */
    int_type getRowDimSparse();

    /**
     * \brief getColDimSparse:
     *        Gibt die col dimension der resultierenden SparseMatrix zurück.
     *
     * @return Gibt die col dimension der resultierenden SparseMatrix zurück.
     *
     */
    int_type getColDimSparse();

    /**
     * \brief clear:
     *        Entfernt die Daten der myDoubleCollectorMatrix.
     */
    void clear();

    /**
     * \brief PushMatrix:
     *        Legt Pointer einer Teilmatrix ab.
     *
     * \param[in] Row Row Index.
     * \param[in] Col Column Index.
     * \param[in] M   Zu schreibende Teilmatrix.
     *
     */
    void PushMatrix(int_type Row, int_type Col, MatrixWrapper<real> * M);

    /**
     * \brief PushMatrixTransposed:
     *        Legt Pointer einer Teilmatrix transponiert ab.
     *
     * \param[in] Row Row Index.
     * \param[in] Col Column Index.
     * \param[in] M   Zu schreibende Teilmatrix.
     *
     */
    void PushMatrixTransposed(int_type Row, int_type Col, MatrixWrapper<real> * M);

    /**
     * \brief PushDiagMatrix:
     *        Legt Pointer einer Teilmatrix ab.
     *
     * \param[in] Row Row Index.
     * \param[in] Col Column Index.
     * \param[in] D   Zu schreibende diagonale Teilmatrix.
     *
     */
    void PushDiagMatrix(int_type Row, int_type Col, DiagMatrixWrapper<real> * D);

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                   protected Membervariablen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

  protected:
    /// Anzahl der Rows.
    int_type nRows;
    /// Anzahl der Column.
    int_type nCols;
    /// Schätzung der Größe der myDoubleSparseMatrix.
    int_type NNz_Estimate;
    /// Flag, der angibt, ob Daten in der myDoubleSparseMatrix symmetrisch abgelegt werden.
    isSym Flag_Sym;

  private:
    /// Abgelegten der Pointer der Matrixelemente:
    std::vector<MatrixWrapper<real> *> AlglinMatrixContainer;
    /// My_std_Vector mit den Row positionen der Teilmatrizen.
    std::vector<int_type> Rows;
    /// My_std_Vector mit den Column positionen der Teilmatrizen.
    std::vector<int_type> Cols;

    /// Abgelegten der Pointer der transponierten Matrixelemente:
    std::vector<MatrixWrapper<real> *> AlglinMatrixContainer_Transposed;
    /// My_std_Vector mit den Rows positionen der transponierten Teilmatrizen.
    std::vector<int_type> Row_Transposed;
    /// My_std_Vector mit den Column positionen der transponierten Teilmatrizen.
    std::vector<int_type> Col_Transposed;

    /// Abgelegten der Pointer der diagonalen Matrixelemente:
    std::vector<DiagMatrixWrapper<real> *> AlglinMatrixContainer_Diag;
    /// My_std_Vector mit den Rows positionen der diagonalen Teilmatrizen.
    std::vector<int_type> Row_Diag;
    /// My_std_Vector mit den Column positionen der diagonalen Teilmatrizen.
    std::vector<int_type> Col_Diag;

    /// Anzahl der Collector Rows.
    int_type N_Coll_Row;
    /// Anzahl der Collector Columns.
    int_type N_Coll_Col;
    /// DimensionenArray der Collector Rows.
    std::vector<int_type> dimRow;
    /// DimensionenArray der Collector Columns.
    std::vector<int_type> dimCol;
    /// Resultierendes gesamt DimensionenArray der Collector Rows.
    std::vector<int_type> ResultingDimRow;
    /// Resultierendes gesamt DimensionenArray der Collector Columns.
    std::vector<int_type> ResultingDimCol;

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    private Memberfunktionen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief internalMatrixPush:
     *        Interne Verarbeitung des MatrixPush.
     *
     * \param[in] Row Row Index.
     * \param[in] Col Column Index.
     * \param[in] M   Zu schreibende Teilmatrix.
     *
     */
    void internalMatrixPush(int_type Row, int_type Col, MatrixWrapper<real> * M);

    /**
     * \brief internalMatrixPushTransposed:
     *        Interne Verarbeitung des MatrixPush.
     *
     * \param[in] Row Row Index.
     * \param[in] Col Column Index.
     * \param[in] M   Zu schreibende Teilmatrix.
     *
     */
    void internalMatrixPushTransposed(int_type Row, int_type Col, MatrixWrapper<real> * M);

    /**
     * \brief internalMatrixPushDiag:
     *        Interne Verarbeitung des MatrixPush.
     *
     * \param[in] Row Row Index.
     * \param[in] Col Column Index.
     * \param[in] D   Zu schreibende diagonale Teilmatrix.
     *
     */
    void internalMatrixPushDiag(int_type Row, int_type Col, DiagMatrixWrapper<real> * D);

    /**
     * \brief internalSetResultingDims:
     *        Belegt die Arrays der resultierenden Dimensionen.
     */
    void internalSetResultingDims();
};

#endif // COLLECTORMATRIX_H
