/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file sparsematrix.h
 * Header Definition der Klasse SparseMatrix.
 *
 * @author A. Huber
 * @since 08.11.2016
 */

#ifndef SPARSEMATRIX_H
#define SPARSEMATRIX_H

#include <IPbasic/IPbasic.h>

using IPbasic::DiagMatrixWrapper;
using IPbasic::int_type;
using IPbasic::MatrixWrapper;

/**
 * \brief Flags zur Abfrage, ob eine Matrix symetrisch ist.
 */
typedef enum
{
    /// Matrix ist symmetrisch.
    SYM,
    /// Matrix ist nicht symmetrisch.
    NOT_SYM,
} isSym;

template <typename real>
class SparseMatrix : public IPbasic::SparseCCOOR<real>
{

    /// Gibt an ob in der Funktion PushVal zwingend Nullelemente in die Sparsematrix geschrieben werden (Default = False).
    bool has_to_push_zeroes;

  protected:
    /// Geschätzter Wert für die Dimension der Sparsematrizen.
    static int_type NNz_Estimate;
    /// Gibt an, ob die Matrix symmetrisch ist.
    isSym Flag_Sym;
    /// Statische Tolerzanz für den real-Wert 0.0.
    static real tol_zero;

  public:
    /**
     * \brief SparseMatrix:
     *        Default Konstruktor.
     *
     * \param[in] Flag_Pr_Sp    Gibt an welche Notation (Fortran oder C, C++) verwendet werden soll.
     * \param[in] Flag_Sym      Gibt an ob die Matrix symmetrisch ist.
     *
     */
    SparseMatrix(bool Flag_Pr_Sp = true, isSym Flag_Sym = NOT_SYM, bool has_to_push_zeroes = false)
        : IPbasic::SparseCCOOR<real>()
        , Flag_Sym(Flag_Sym)
    {
        this->has_to_push_zeroes = has_to_push_zeroes;
        this->fortran_indexing   = Flag_Pr_Sp;
        this->init(0, 0, SparseMatrix<real>::NNz_Estimate);
    }

    /**
     * \brief SparseMatrix:
     *        Konstuktor allokiert Arrays.
     *
     * \param[in] NNz_Estimate  Maximale Anzahl von Einträgen der Sparsematrix.
     * \param[in] Flag_Pr_Sp          Gibt an welche Notation (Fortran oder C, C++) verwendet werden soll.
     * \param[in] Flag_Sym            Gibt an ob die Matrix symmetrisch ist.
     *
     */
    SparseMatrix(int_type NNz_Estimate, bool Flag_Pr_Sp = true, isSym Flag_Sym = NOT_SYM, bool has_to_push_zeroes = false)
        : IPbasic::SparseCCOOR<real>()
        , Flag_Sym(Flag_Sym)
    {
        this->has_to_push_zeroes = has_to_push_zeroes;
        this->fortran_indexing   = Flag_Pr_Sp;
        this->init(0, 0, NNz_Estimate);
    }

    /**
     * \brief SparseMatrix:
     *        Konstuktor allokiert Arrays und belegt aktuelle row dimension.
     *
     * \param[in] dim_i                 Vorgegbene aktuelle row dimension.
     * \param[in] dim_j                 Vorgegbene aktuelle col dimension.
     * \param[in] NNz_Estimate       Maximale Anzahl von Einträgen der Sparsematrix.
     * \param[in] Flag_Pr_Sp            Gibt an welche Notation (Fortran oder C, C++) verwendet werden soll.
     * \param[in] Flag_Sym              Gibt an ob die Matrix symmetrisch ist.
     *
     */
    SparseMatrix(int_type dim_i, int_type dim_j, int_type NNz_Estimate, bool Flag_Pr_Sp = true, isSym Flag_Sym = NOT_SYM, bool has_to_push_zeroes = false)
        : IPbasic::SparseCCOOR<real>()
        , Flag_Sym(Flag_Sym)
    {
        this->has_to_push_zeroes = has_to_push_zeroes;
        this->fortran_indexing   = Flag_Pr_Sp;
        this->init(dim_i, dim_j, NNz_Estimate);
    }

    /**
     * \brief SparseMatrix:
     *        Konstruktor allokiert die Sparsematrix und belegt kopiert alle Werte.
     *
     * \param[in] Copy  Zu kopierende Sparsematrix.
     *
     */
    SparseMatrix(SparseMatrix const & Copy)
        : IPbasic::SparseCCOOR<real>()
    {
        // Kopieren der Flags:
        this->has_to_push_zeroes = Copy.has_to_push_zeroes;
        this->fortran_indexing   = Copy.fortran_indexing;
        this->Flag_Sym           = Copy.Flag_Sym;
        // Kopieren der Dimensionen:
        this->nnz   = Copy.nnz;
        this->nRows = Copy.nRows;
        this->nCols = Copy.nCols;
        // Kopieren der dynamischen Vektoren:
        this->rows = Copy.rows;
        this->cols = Copy.cols;
        this->vals = Copy.vals;
    }

    /**
     * \brief ~SparseMatrix:
     *        Virtueller Destruktor der Klasse SparseMatrix.
     */
    virtual ~SparseMatrix()
    {
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                     public Memberfunktionen                                                         //
    //                                                           Overloaded                                                                //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    /*!
     * \brief gemv:
     *        Calls the blas-Routine (\f$y = \alpha \mathrm{op}(A) x + y\f$).
     *
     * \param[in]     alpha         Scalar \f$\alpha \f$.
     * \param[in]     DimX          Dimension of the vector x.
     * \param[in]     x             Vector x.
     * \param[in]     incX          Stride of vector x.
     * \param[in]     DimY          Dimension of the vector y.
     * \param[in,out] y             In-/Output vector y.
     * \param[in]     incY          Stride of vector y.
     *
    \*/

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                     public Memberfunktionen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief PushVal:
     *        Legt einen Wert in der SparseMatrix ab.
     *
     * \param[in] row Row index.
     * \param[in] col Column Index.
     * \param[in] val Wert.
     *
     */
    void PushVal(int_type row, int_type col, real val);

    /**
     * \brief PushMatrix:
     *        Legt eine Teilmatrix in der SparseMatrix ab.
     *
     * \param[in] row Row index.
     * \param[in] col Column Index.
     * \param[in] M   Zu schreibende Teilmatrix.
     *
     */
    void PushMatrix(int_type row, int_type col, MatrixWrapper<real> const & M);

    /**
     * \brief PushMatrixTransposed:
     *        Legt eine Teilmatrix transponiert in der SparseMatrix ab.
     *
     * \param[in] row Row index.
     * \param[in] col Column Index.
     * \param[in] M   Zu schreibende Teilmatrix.
     *
     */
    void PushMatrixTransposed(int_type row, int_type col, MatrixWrapper<real> const & M);

    /**
     * \brief getN_Aktuell:
     *        Gibt die aktuelle Dimension zurück.
     *
     * @return Wert der aktuellen Dimension.
     *
     */
    int_type getN_Aktuell();

    /**
     * \brief copyArrays:
     *        Kopiert alle Datenarrays.
     *
     * \param[in]  N_Aktuell     Aktuelle Dimension.
     * \param[out] i             Array für rownpositionen der Einträge der Matrix.
     * \param[out] j             Array für Column positionen der Einträge der Matrix.
     * \param[out] val           Array für Werte der Einträge der Matrix.
     */
    void export_data(int_type N_Aktuell, int_type i[], int_type j[], real val[]);

    /**
     * \brief FillDiag:
     *        Belegt Diagonalelemente mit 0 falls an dieser Stelle kein Wert vorhanden ist.
     */
    void FillDiag();

    /**
     * \brief GerschgorinRegularization:
     *        Regularisierung einer Hessematrix durch Abschätzung der Eigenwerte mit Gerschgorin-Kreisen.
     */
    void GerschgorinRegularization();

    /**
     * \brief compareDim:
     *        Vergleicht die Dimensionen von zwei Sparsematrizen.
     *
     * \param[in] B     Zu vergleichende SparseMatrix.
     *
     * @return True, falls Dimensionen gleich. False, falls Dimension nicht gleich.
     *
     */
    bool compareDim(SparseMatrix const & B);

    /**
     * \brief setZero:
     *        Entfernt die Daten der Sparsematrix.
     */
    void setZero()
    {
        this->nnz = 0;
        this->rows.clear();
        this->cols.clear();
        this->vals.clear();
    }

    /**
     * \brief isSymmetric:
     *        Gibt an ob eine Matrix symmetrisch abgespeichert wurde.
     *
     * @return True, falls SparseMatrix symmetrisch. False, falls SparseMatrix nicht symmetrisch.
     *
     */
    bool isSymmetric() const;

    /**
     * \brief SymmetricalFilling:
     *        Füllt die symmetrischen Anteile in der Matrix auf.
     */
    void SymmetricalFilling();

    /**
     * \brief printFile:
     *        Schreibt SparseMatrix in ein File.
     *
     * \param[in] str     Name des Files.
     *
     */
    void printFile(std::string const & str);

    /**
     * \brief setDimensionen:
     *        Dimensionier die Matrix neu.
     *
     * \param[in] Dim_i     Neue Dimension der Rows.
     * \param[in] Dim_j     Neue Dimension der Columns.
     *
     */
    void setDimension(int_type Dim_i, int_type Dim_j)
    {
        this->setZero();
        this->nRows = Dim_i;
        this->nCols = Dim_j;
    }

    /**
     * \brief sortSparseMatrix:
     *        Calculated die aus den aktuellen Daten resultierende SparseMatrix (Umsortiert).
     *
     * \param[in] DimROW    Dimension von SortROW.
     * \param[in] SortROW   Sortierungsarray für die rown.
     * \param[in] DimCOL    Dimension von SortCOL.
     * \param[in] SortCOL   Sortierungsarray für die Column.
     *
     */
    void sortSparseMatrix(int_type DimROW, int_type SortROW[], int_type DimCOL, int_type SortCOL[]);

    /**
     * \brief SetHasToPushZeros:
     *        Sets the HasToPushZeros flag in the sparse matrix.
     *
     * \param[in] HasToPushZeros    Flag to tell wether zeroes should be pushed.
     *
     */
    void SetHasToPushZeros(bool HasToPushZeros);

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                  Externe Memberoperatoren                                                           //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief operator =
     *        Zuweisungsoperator für Objekte.
     *
     * \param[in] B     Zu übernehmende SparseMatrix.
     *
     * @return Überschriebene SparseMatrix.
     *
     */
    SparseMatrix & operator=(SparseMatrix const & B);

    /**
     * \brief toOStream:
     *        Gibt die SparseMatrix aus.
     *
     * \param[in,out] out    Übergebener std::ostream.
     *
     */
    void toOStream(std::ostream & out);

  private:
    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                     private Memberfunktionen                                                        //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief init:
     *        Initialisiert die Sparsematrix.
     *
     * \param[in] dim_i                 Aktuelle row dimension.
     * \param[in] dim_j                 Aktuelle col dimension.
     * \param[in] NNz_Estimate       Geschätzter Wert für die Dimension der Sparsematrizen.
     *
     */
    void init(int_type dim_i, int_type dim_j, int_type NNz_Estimate)
    {
        this->IPbasic::SparseCCOOR<real>::init(dim_i, dim_j, NNz_Estimate);
        this->has_to_push_zeroes = true;
    }

    /**
     * \brief PushBackNote:
     *        Legt ein Wertepaar entsrechend der Vorgegebenen Notation ab.
     *
     * \param[in] row Row index.
     * \param[in] col Column Index.
     * \param[in] val Wert.
     *
     */
    void PushBackNote(int_type row, int_type col, real val);
};

#endif // SPARSEMATRIX_H
