/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file problem.h
 * Header definition of the class Problem.
 *
 * @author A. Huber
 * @since 21.04.2016
 */

#ifndef PROBLEM_H
#define PROBLEM_H

#include <IPbasic/IPbasic.h>
#include <string>

namespace OCPbasic
{

    using IPbasic::DiagMatrixWrapper;
    using IPbasic::int_type;
    using IPbasic::MatrixWrapper;

    template <typename real>
    class Problem
    {
      protected:
        /**
         * \brief Problem:
         *        Contructor of the class Problem.
         *
         * \param[in] Ny         Dimension of states.
         * \param[in] Nu         Dimension of controls.
         * \param[in] Np         Dimension of parameters.
         * \param[in] N_NLCSTR   Number of nonlinear constraints.
         * \param[in] N_BOUND    Number of boundary constraints.
         *
         */
        Problem(int_type Ny, int_type Nu, int_type Np, int_type N_NLCSTR, int_type N_BOUND);

      public:
        /**
         * \brief ~Problem:
         *        Virtual destructor of the class Problem.
         */
        virtual ~Problem();

        /// Dimension of states.
        int_type Ny;
        /// Dimension of controls.
        int_type Nu;
        /// Dimension of parameters.
        int_type Np;
        /// Number of nonlinear constraints.
        int_type N_NLCSTR;
        /// Number of boundary constraints.
        int_type N_BOUND;
        /// Dimension of the target function. (= 1).
        int_type N_VARPHI;
        /// Lower bounds of the nonlinear constraints.
        real * g_min;
        /// Upper bounds of the nonlinear constraints.
        real * g_max;
        /// Lower bounds of the boundary constraints.
        real * psi_min;
        /// Upper bounds of the boundary constraints.
        real * psi_max;
        /// Lower bounds of the states.
        real * y_min;
        /// Upper bounds of the states.
        real * y_max;
        /// Lower bounds of the controls.
        real * u_min;
        /// Upper bounds of the controls.
        real * u_max;
        /// Lower bounds of the parameters.
        real * p_min;
        /// Upper bounds of the parameters.
        real * p_max;

        /// Value for \f$\infty\f$:
        static const real Inf;

        /**
         * \brief ODE:
         *        Righthand side of the ODE.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] f      Righthand side of the ODE as a vector.
         *
         */
        virtual void ODE(real t, real const y[], real const u[], real const p[], real f[]) const = 0;

        /**
         * \brief NLCSTR:
         *        Nonlinear constaints.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] gg     Nonlinear constaints as a vector.
         *
         */
        virtual void NLCSTR(real t, real const y[], real const u[], real const p[], real gg[]) const = 0;

        /**
         * \brief BOUND:
         *        Boundary constaints.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] psi    Boundary constaints as a vector.
         *
         */
        virtual void BOUND(real t0, real tF, real const yF[], real const y0[], real const p[], real psi[]) const = 0;

        /**
         * \brief VARPHI:
         *        Target or objective function.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] phi    Target or objective function (vector of size 1).
         *
         */
        virtual void VARPHI(real t0, real tF, real const yF[], real const y0[], real const p[], real phi[]) const = 0;

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                    Erste Ableitungen                                                                //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                             Rechte Seite der DGL                                                    //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief D_ODE_X:
         *        First derivative or jacobian of the ODEs towards \f$y\f$.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the ODEs towards \f$y\f$.
         *
         */
        virtual void D_ODE_X(real t, real const y[], real const u[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_ODE_U:
         *        First derivative or jacobian of the ODEs towards \f$u\f$.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the ODEs towards \f$u\f$.
         *
         */
        virtual void D_ODE_U(real t, real const y[], real const u[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_ODE_P:
         *        First derivative or jacobian of the ODEs towards \f$p\f$.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the ODEs towards \f$p\f$.
         *
         */
        virtual void D_ODE_P(real t, real const y[], real const u[], real const p[], MatrixWrapper<real> & S_OUT);

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                        Nichtlineare Nebenbedingungen                                                //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief D_NLCSTR_X:
         *        First derivative or jacobian of the nonlinear constraints towards \f$y\f$.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the nonlinear constraints towards \f$y\f$.
         *
         */
        virtual void D_NLCSTR_X(real t, real const y[], real const u[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_NLCSTR_U:
         *        First derivative or jacobian of the nonlinear constraints towards \f$u\f$.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the nonlinear constraints towards \f$u\f$.
         *
         */
        virtual void D_NLCSTR_U(real t, real const y[], real const u[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_NLCSTR_P:
         *        First derivative or jacobian of the nonlinear constraints towards \f$p\f$.
         *
         * \param[in]  t      Timepoint \f$t\f$.
         * \param[in]  y      States at \f$y(t)\f$.
         * \param[in]  u      Controls at \f$u(t)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the nonlinear constraints towards \f$p\f$.
         *
         */
        virtual void D_NLCSTR_P(real t, real const y[], real const u[], real const p[], MatrixWrapper<real> & S_OUT);

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                               Randbedingungen                                                       //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief D_BOUND_XF:
         *        First derivative or jacobian of the boundary constraints towards \f$y_f\f$.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the boundary constraints towards \f$y_f\f$.
         *
         */
        virtual void D_BOUND_XF(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_BOUND_X0:
         *        First derivative or jacobian of the boundary constraints towards \f$y_0\f$.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the boundary constraints towards \f$y_0\f$.
         *
         */
        virtual void D_BOUND_X0(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_BOUND_P:
         *        First derivative or jacobian of the boundary constraints towards \f$p\f$.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the boundary constraints towards \f$p\f$.
         *
         */
        virtual void D_BOUND_P(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                                 Zielfunktion                                                        //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief D_VARPHI_XF:
         *        First derivative or jacobian of the target function towards \f$y_f\f$.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the target function towards \f$y_f\f$.
         *
         */
        virtual void D_VARPHI_XF(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_VARPHI_X0:
         *        First derivative or jacobian of the target function towards \f$y_0\f$.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the target function towards \f$y_0\f$.
         *
         */
        virtual void D_VARPHI_X0(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief D_VARPHI_P:
         *        First derivative or jacobian of the target function towards \f$p\f$.
         *
         * \param[in]  t0     Start timepoint \f$t_0\f$.
         * \param[in]  tF     End timepoint \f$t_f\f$.
         * \param[in]  yF     States at \f$y(t_f)\f$.
         * \param[in]  y0     States at \f$y(t_0)\f$.
         * \param[in]  p      Parameters \f$p\f$.
         * \param[out] S_OUT  Jacobian of the target function towards \f$p\f$.
         *
         */
        virtual void D_VARPHI_P(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                    Zweite Ableitungen                                                               //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                             Rechte Seite der DGL                                                    //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief L_DD_ODE_XX:
         *        Second derivative or hessian of the ODEs towards \f$y,y\f$, multiplied by the
         *        lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the ODEs towards \f$y,y\f$, multiplied by the
         *                          lagrangian multipliers.
         *
         */
        virtual void L_DD_ODE_XX(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_ODE_UU:
         *        Second derivative or hessian of the ODEs towards \f$u,u\f$, multiplied by the
         *        lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the ODEs towards \f$u,u\f$, multiplied by the
         *                          lagrangian multipliers.
         *
         */
        virtual void L_DD_ODE_UU(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_ODE_PP:
         *        Second derivative or hessian of the ODEs towards \f$p,p\f$, multiplied by the
         *        lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the ODEs towards \f$p,p\f$, multiplied by the
         *                          lagrangian multipliers.
         *
         */
        virtual void L_DD_ODE_PP(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_ODE_XU:
         *        Second derivative or hessian of the ODEs towards \f$y,u\f$, multiplied by the
         *        lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the ODEs towards \f$y,u\f$, multiplied by the
         *                          lagrangian multipliers.
         *
         */
        virtual void L_DD_ODE_XU(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_ODE_XP:
         *        Second derivative or hessian of the ODEs towards \f$y,p\f$, multiplied by the
         *        lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the ODEs towards \f$y,p\f$, multiplied by the
         *                          lagrangian multipliers.
         *
         */
        virtual void L_DD_ODE_XP(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_ODE_UP:
         *        Second derivative or hessian of the ODEs towards \f$u,p\f$, multiplied by the
         *        lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the ODEs towards \f$u,p\f$, multiplied by the
         *                          lagrangian multipliers.
         *
         */
        virtual void L_DD_ODE_UP(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                        Nichtlineare Nebenbedingungen                                                //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief L_DD_NLCSTR_XX:
         *        Second derivative or hessian of the nonlinear constraints towards \f$y,y\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$y,y\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_NLCSTR_XX(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_NLCSTR_UU:
         *        Second derivative or hessian of the nonlinear constraints towards \f$u,u\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$u,u\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_NLCSTR_UU(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_NLCSTR_PP:
         *        Second derivative or hessian of the nonlinear constraints towards \f$p,p\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$p,p\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_NLCSTR_PP(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_NLCSTR_XU:
         *        Second derivative or hessian of the nonlinear constraints towards \f$y,u\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$y,u\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_NLCSTR_XU(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_NLCSTR_XP:
         *        Second derivative or hessian of the nonlinear constraints towards \f$y,p\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$y,p\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_NLCSTR_XP(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_NLCSTR_UP:
         *        Second derivative or hessian of the nonlinear constraints towards \f$u,p\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t            Timepoint \f$t\f$.
         * \param[in]  y            States at \f$y(t)\f$.
         * \param[in]  u            Controls at \f$u(t)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$u,p\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_NLCSTR_UP(real t, real const y[], real const u[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                               Randbedingungen                                                       //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief L_DD_BOUND_XFXF:
         *        Second derivative or hessian of the boundary constraints towards \f$y_f,y_f\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_f,y_f\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_BOUND_XFXF(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_BOUND_X0X0:
         *        Second derivative or hessian of the boundary constraints towards \f$y_0,y_0\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_0,y_0\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_BOUND_X0X0(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_BOUND_PP:
         *        Second derivative or hessian of the boundary constraints towards \f$p,p\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the boundary constraints towards \f$p,p\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_BOUND_PP(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_BOUND_XFX0:
         *        Second derivative or hessian of the boundary constraints towards \f$y_f,y_0\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_f,y_0\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_BOUND_XFX0(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_BOUND_XFP:
         *        Second derivative or hessian of the boundary constraints towards \f$y_f,p\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_f,p\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_BOUND_XFP(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_BOUND_X0P:
         *        Second derivative or hessian of the boundary constraints towards \f$y_0,p\f$,
         *        multiplied by the lagrangian multipliers.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
         * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_0,p\f$,
         *                          multiplied by the lagrangian multipliers.
         *
         */
        virtual void L_DD_BOUND_X0P(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], MatrixWrapper<real> & S_OUT);

        //---------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                     //
        //                                                 Zielfunktion                                                        //
        //                                                                                                                     //
        //---------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief L_DD_VARPHI_XFXF:
         *        Second derivative or hessian of the target fuction towards \f$y_f,y_f\f$.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[out] S_OUT        Hessian of the target fuction towards \f$y_f,y_f\f$.
         *
         */
        virtual void L_DD_VARPHI_XFXF(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_VARPHI_X0X0:
         *        Second derivative or hessian of the target fuction towards \f$y_0,y_0\f$.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[out] S_OUT        Hessian of the target fuction towards \f$y_0,y_0\f$.
         *
         */
        virtual void L_DD_VARPHI_X0X0(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_VARPHI_PP:
         *        Second derivative or hessian of the target fuction towards \f$p,p\f$.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[out] S_OUT        Hessian of the target fuction towards \f$p,p\f$.
         *
         */
        virtual void L_DD_VARPHI_PP(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_VARPHI_XFX0:
         *        Second derivative or hessian of the target fuction towards \f$y_f,y_0\f$.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[out] S_OUT        Hessian of the target fuction towards \f$y_f,y_0\f$.
         *
         */
        virtual void L_DD_VARPHI_XFX0(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_VARPHI_XFP:
         *        Second derivative or hessian of the target fuction towards \f$y_f,p\f$.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[out] S_OUT        Hessian of the target fuction towards \f$y_f,p\f$.
         *
         */
        virtual void L_DD_VARPHI_XFP(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

        /**
         * \brief L_DD_VARPHI_X0P:
         *        Second derivative or hessian of the target fuction towards \f$y_0,p\f$.
         *
         * \param[in]  t0           Start timepoint \f$t_0\f$.
         * \param[in]  tF           End timepoint \f$t_f\f$.
         * \param[in]  yF           States at \f$y(t_f)\f$.
         * \param[in]  y0           States at \f$y(t_0)\f$.
         * \param[in]  p            Parameters \f$p\f$.
         * \param[out] S_OUT        Hessian of the target fuction towards \f$y_0,p\f$.
         *
         */
        virtual void L_DD_VARPHI_X0P(real t0, real tF, real const yF[], real const y0[], real const p[], MatrixWrapper<real> & S_OUT);

      protected:
        /// Tolerance of central differnces.
        real tolCentralDiff;

      private:
        /// Working real space 1.
        real * workspace1;
        /// Working real space vector 1.
        real * workspaceVector1;
        /// Working real space vector 2.
        real * workspaceVector2;
        /// Working real space vector 3.
        real * workspaceVector3;

        /**
         * \brief Set_Default_Bounds:
         *        Sets all bounds to -Inf or Inf.
         */
        void Set_Default_Bounds();

        /**
         * \brief multMatrixVector:
         *        Matrix Vector multiplication.
         *
         * \param[in]  matrix       Matrix M.
         * \param[in]  vector       Vector v.
         * \param[out] result       Result M*v.
         *
         */
        void multMatrixVector(const MatrixWrapper<real> & matrix, const real * vector, real * result);
    };

} // namespace OCPbasic

#endif // PROBLEM_H
