/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file solution.h
 * Header definition of the class Solution.
 *
 * @author A. Huber
 * @since 07.11.2018
 */

#ifndef SOLUTION_H
#define SOLUTION_H

#include <IPbasic/IPbasic.h>
#include <string>

namespace OCPbasic
{

    using IPbasic::int_type;

    template <typename real>
    class Problem;

    template <typename real>
    class Solution
    {
      public:
        /**
         * \brief Solution:
         *        Contructor of the class Solution.
         *
         * \param[in] dim_X     Dimension of the statevectors.
         * \param[in] dim_U     Dimension of the controlvectors.
         * \param[in] dim_P     Dimension of the parametervector.
         * \param[in] grid_X    Number of state gridpoints.
         * \param[in] grid_U    Number of control gridpoints.
         *
         */
        Solution(int_type dim_X, int_type dim_U, int_type dim_P, int_type grid_X, int_type grid_U);

        /**
         * \brief ~Solution:
         *        Virtual destructor of the class Solution.
         */
        virtual ~Solution();

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                     Public Membervariablen                                                          //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /// Dimension of the statevector.
        int_type dim_X;
        /// Dimension of the controlvector.
        int_type dim_U;
        /// Dimension of the parametervector.
        int_type dim_P;
        /// Number of state gridpoints.
        int_type grid_X;
        /// Number of control gridpoints.
        int_type grid_U;
        /// Statesvectors of the solution.
        real ** X;
        /// Controlvectors of the solution.
        real ** U;
        /// Parametervector of the solution.
        real * P;
        /// Timegrid of the states.
        real * t_X;
        /// Timegrid of the controls.
        real * t_U;
        /// Tell wether the solution was already set.
        bool isSet;
        /// Tells if the controlls are piecewise constant.
        bool isStKonst;
        /// Tells if MaxIter was reached.
        bool isMaxIter;

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                     Public Memberfunktionen                                                         //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief PrintFile_X:
         *        Writes the statevectors into a file.
         *
         * \param[in] file      Name of the File.
         *
         */
        virtual void PrintFile_X(std::string const & file);

        /**
         * \brief PrintFile_U:
         *        Writes the controlvectors into a file.
         *
         * \param[in] file      Name of the File.
         *
         */
        virtual void PrintFile_U(std::string const & file);

        /**
         * \brief PrintFile_P:
         *        Writes the parametervector into a file.
         *
         * \param[in] file      Name of the File.
         *
         */
        virtual void PrintFile_P(std::string const & file);

        /**
         * \brief PrintFile_XUP:
         *        Writes the statevectors, controlvectors and parametervector into a file.
         *
         * \param[in] file      Name of the File.
         *
         */
        virtual void PrintFile_XUP(std::string const & file);

        /**
         * \brief PrintFile_TimeScale_X:
         *        Writes the statevectors into a file with a scaled timearray.
         *
         * \param[in] file      Name of the File.
         * \param[in] scale     Scaling factor of the timearray.
         *
         */
        virtual void PrintFile_TimeScale_X(std::string const & file, real scale);

        /**
         * \brief PrintFile_TimeScale_U:
         *        Writes the controlvectors into a file with a scaled timearray.
         *
         * \param[in] file      Name of the File.
         * \param[in] scale     Scaling factor of the timearray.
         *
         */
        virtual void PrintFile_TimeScale_U(std::string const & file, real scale);

        /**
         * \brief PrintFile_TimeScale_XUP:
         *        Writes the statevectors, controlvectors and parametervector into a file with a scaled timearray.
         *
         * \param[in] file      Name of the File.
         * \param[in] scale     Scaling factor of the timearray.
         *
         */
        virtual void PrintFile_TimeScale_XUP(std::string const & file, real scale);

        /**
         * \brief integrateStartSolutionEuler:
         *        Integrates a startsolution with euler integration.
         *
         * \param[in] OCP   Optimal control problem.
         * \param[in] t     Vector with timegrid \f$t_i\f$.
         * \param[in] y0    Starting state \f$y_0\f$.
         * \param[in] u     Integration control \f$u\f$.
         * \param[in] p0    Startparameter \f$p_0\f$.
         *
         */
        void integrateStartSolutionEuler(Problem<real> const * OCP, real const t[], real const y0[], real const u[], real const p0[] = nullptr);

        /**
         * \brief integrateStartSolutionEuler:
         *        Integrates a startsolution with euler integration.
         *
         * \param[in] OCP   Optimal control problem.
         * \param[in] h     Equidistant stepsize in time \f$h\f$.
         * \param[in] y0    Starting state \f$y_0\f$.
         * \param[in] u     Integration control \f$u\f$.
         * \param[in] p0    Startparameter \f$p_0\f$.
         *
         */
        void integrateStartSolutionEuler(Problem<real> const * OCP, real h, real const y0[], real const u[], real const p0[] = nullptr);

      private:
        /**
         * \brief testFileType:
         *        Check if a file path has correct file type.
         *
         * \param[in] FilePath      File path.
         * \param[in] Type          File type.
         *
         * @return True, if ile path has correct file type.
         *
         */
        bool testFileType(std::string FilePath, std::string Type);
    };

} // namespace OCPbasic

#endif // SOLUTION_H
